#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Library of common forms for phase equilibrium constraints
"""
from pyomo.environ import ConcreteModel, Var, units as pyunits

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx

from idaes.models.properties.modular_properties.phase_equil.forms import *
import pytest


# Dummy EoS to use for fugacity calls
class DummyEoS(object):
    def common(self):
        pass

    def build_parameters(b):
        pass

    def fug_phase_comp_eq(b, p, j, pp):
        return b.x[p, j]

    def log_fug_phase_comp_eq(b, p, j, pp):
        return 42 * b.x[p, j]


@pytest.mark.unit
def test_fugacity():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3},
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {"equation_of_state": DummyEoS},
            "Vap": {"equation_of_state": DummyEoS},
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    )

    assert str(fugacity.return_expression(m, "Vap", "Liq", "H2O")) == str(
        m.x["Vap", "H2O"] == m.x["Liq", "H2O"]
    )


@pytest.mark.unit
def test_log_fugacity():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3},
                "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            }
        },
        phases={
            "Liq": {"equation_of_state": DummyEoS},
            "Vap": {"equation_of_state": DummyEoS},
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    )

    assert str(log_fugacity.return_expression(m, "Vap", "Liq", "H2O")) == str(
        42 * m.x["Vap", "H2O"] == 42 * m.x["Liq", "H2O"]
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
import re

from pyomo.environ import ConcreteModel, Set, TransformationFactory, units
from pyomo.dae import ContinuousSet
from pyomo.network import Arc

from idaes.core import (
    FlowsheetBlockData,
    FlowsheetBlock,
    UnitModelBlock,
    declare_process_block_class,
    useDefault,
)
from idaes.core.base import flowsheet_model
from idaes.core.util.exceptions import DynamicError
from idaes.core.util.testing import PhysicalParameterTestBlock

from idaes.models.unit_models import Heater


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(FlowsheetBlockData, self).build()


class TestConfig(object):
    @pytest.fixture()
    def model(self):
        m = ConcreteModel()
        m.fs = Flowsheet()

        return m

    @pytest.mark.unit
    def test_config(self, model):
        assert len(model.fs.config) == 5
        assert model.fs.config.dynamic is useDefault
        assert model.fs.config.time is None
        assert model.fs.config.time_set == [0]
        assert model.fs.config.default_property_package is None
        assert model.fs.config.time_units is None

    @pytest.mark.unit
    def test_config_validation_dynamic(self, model):
        # Test validation of dynamic argument
        model.fs.config.dynamic = False
        assert model.fs.config.dynamic is False

        model.fs.config.dynamic = True
        assert model.fs.config.dynamic is True

        with pytest.raises(ValueError):
            model.fs.config.dynamic = "foo"
        with pytest.raises(ValueError):
            model.fs.config.dynamic = 2
        with pytest.raises(ValueError):
            model.fs.config.dynamic = 2.0
        with pytest.raises(ValueError):
            model.fs.config.dynamic = [1]
        with pytest.raises(ValueError):
            model.fs.config.dynamic = {"foo": 1}

    @pytest.mark.unit
    def test_config_validation_time(self, model):
        # Test validation of time argument
        model.test_set = Set(initialize=[0, 1, 2])
        model.test_contset = ContinuousSet(bounds=[0, 1])

        model.fs.config.time = model.test_set
        assert model.fs.config.time == model.test_set

        model.fs.config.time = model.test_contset
        assert model.fs.config.time == model.test_contset

        with pytest.raises(ValueError):
            model.fs.config.time = "foo"
        with pytest.raises(ValueError):
            model.fs.config.time = 1
        with pytest.raises(ValueError):
            model.fs.config.time = 2.0
        with pytest.raises(ValueError):
            model.fs.config.time = [1]
        with pytest.raises(ValueError):
            model.fs.config.time = {"foo": 1}

    @pytest.mark.unit
    def test_config_validation_time_set(self, model):
        # Test validation of time_set argument
        model.fs.config.time_set = [1, 2, 3]
        assert model.fs.config.time_set == [1, 2, 3]
        model.fs.config.time_set = 5
        assert model.fs.config.time_set == [5]
        model.fs.config.time_set = 2.0
        assert model.fs.config.time_set == [2.0]

        with pytest.raises(ValueError):
            model.fs.config.time_set = "foo"  # invalid str
        with pytest.raises(ValueError):
            model.fs.config.time_set = {"a": 2.0}  # invalid dict

    @pytest.mark.unit
    def test_config_validation_default_property_package(self, model):
        # Test default_property_package attribute
        model.fs.p = PhysicalParameterTestBlock()

        model.fs.config.default_property_package = model.fs.p

        # Test default_property_package - invalid values
        with pytest.raises(ValueError):
            model.fs.config.default_property_package = "foo"  # invalid str
        with pytest.raises(ValueError):
            model.fs.config.default_property_package = 5  # invalid int
        with pytest.raises(ValueError):
            model.fs.config.default_property_package = 2.0  # invalid float
        with pytest.raises(ValueError):
            model.fs.config.default_property_package = [2.0]  # invalid list
        with pytest.raises(ValueError):
            model.fs.config.default_property_package = {"a": 2.0}  # invalid dict


class TestBuild(object):
    # Test that build method works for all combinations of config arguments
    @pytest.mark.unit
    def test_is_flowsheet(self):
        # Test that flowsheet has is_flowsheet method and that it returns True
        m = ConcreteModel()
        m.fs = FlowsheetBlock()

        assert hasattr(m.fs, "is_flowsheet")
        assert m.fs.is_flowsheet()

    @pytest.mark.unit
    def test_flowsheet(self):
        # Should return None
        m = ConcreteModel()
        m.fs = FlowsheetBlock()

        assert m.fs.flowsheet() is None

    @pytest.mark.unit
    def test_default(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock()

        assert m.fs.config.dynamic is False
        assert isinstance(m.fs.time, Set)
        assert m.fs.time == [0]
        assert m.fs.config.time is m.fs.time
        assert m.fs.time_units is None

    @pytest.mark.unit
    def test_ss_default(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        assert m.fs.config.dynamic is False
        assert isinstance(m.fs.time, Set)
        assert m.fs.time == [0]
        assert m.fs.config.time is m.fs.time
        assert m.fs.time_units is None

    @pytest.mark.unit
    def test_ss_time_set(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False, time_set=[1, 2, 3])

        assert m.fs.config.dynamic is False
        assert isinstance(m.fs.time, Set)
        for t in m.fs.time:
            assert t in [1, 2, 3]
        assert len(m.fs.time) == 3
        assert m.fs.config.time is m.fs.time
        assert m.fs.time_units is None

    @pytest.mark.unit
    def test_dynamic_default(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)

        assert m.fs.config.dynamic is True
        assert isinstance(m.fs.time, ContinuousSet)
        for t in m.fs.time:
            assert t in [0, 1]
        assert m.fs.config.time is m.fs.time
        assert m.fs.time_units == units.s

    @pytest.mark.unit
    def test_dynamic_time_set(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[1, 2], time_units=units.s)

        assert m.fs.config.dynamic is True
        assert isinstance(m.fs.time, ContinuousSet)
        for t in m.fs.time:
            assert t in [1, 2]
        assert m.fs.config.time is m.fs.time
        assert m.fs.time_units == units.s

    @pytest.mark.unit
    def test_dynamic_time_set_invalid(self):
        m = ConcreteModel()

        with pytest.raises(
            DynamicError,
            match=re.escape(
                "Flowsheet provided with invalid "
                "time_set attribute - must have at "
                "least two values (start and end)."
            ),
        ):
            m.fs = FlowsheetBlock(dynamic=True, time_set=1, time_units=units.s)

    @pytest.mark.unit
    def test_ss_external_time(self):
        m = ConcreteModel()
        m.s = Set(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=False, time=m.s)

        assert m.fs.config.dynamic is False
        assert m.fs.config.time is m.s
        assert m.fs.time is m.s
        assert m.fs.time_units is None

    @pytest.mark.unit
    def test_dynamic_external_time_continuous(self):
        m = ConcreteModel()
        m.s = ContinuousSet(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=False, time=m.s)

        assert m.fs.config.dynamic is False
        assert m.fs.config.time is m.s
        assert m.fs.time is m.s
        assert m.fs.time_units is None

    @pytest.mark.unit
    def test_dynamic_external_time(self):
        m = ConcreteModel()
        m.s = ContinuousSet(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=True, time=m.s, time_units=units.s)

        assert m.fs.config.dynamic is True
        assert m.fs.config.time is m.s
        assert m.fs.time is m.s
        assert m.fs.time_units == units.s

    @pytest.mark.unit
    def test_dynamic_external_time_invalid(self):
        m = ConcreteModel()
        m.s = Set(initialize=[4, 5])

        with pytest.raises(DynamicError):
            m.fs = FlowsheetBlock(dynamic=True, time=m.s, time_units=units.s)

    @pytest.mark.unit
    def test_ss_external_time_and_time_set(self):
        # Should ignore time set
        m = ConcreteModel()
        m.s = Set(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=False, time=m.s, time_set=[1, 2])

        assert m.fs.config.dynamic is False
        assert m.fs.config.time is m.s
        assert m.fs.time is m.s
        assert m.fs.time_units is None

    @pytest.mark.unit
    def test_dynamic_external_time_and_time_set(self):
        # Should ignore time set
        m = ConcreteModel()
        m.s = ContinuousSet(initialize=[4, 5])
        m.fs = FlowsheetBlock(
            dynamic=True, time=m.s, time_set=[1, 2], time_units=units.s
        )

        assert m.fs.config.dynamic is True
        assert m.fs.config.time is m.s
        assert m.fs.time is m.s
        assert m.fs.time_units == units.s

    @pytest.mark.unit
    def test_time_units_ss(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False, time_units=units.s)

        assert m.fs.time_units is units.s

    @pytest.mark.unit
    def test_time_units_dynamic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)

        assert m.fs.time_units is units.s

    @pytest.mark.unit
    def test_time_units_external(self):
        # Should ignore time set
        m = ConcreteModel()
        m.s = ContinuousSet(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=True, time=m.s, time_units=units.s)

        assert m.fs.time_units is units.s


class TestSubFlowsheetBuild(object):
    # Test that build method works with nested flowsheets
    @pytest.mark.unit
    def test_flowsheet(self):
        # With nested flowsheet, flowsheet should return parent
        m = ConcreteModel()
        m.fs = FlowsheetBlock()
        m.fs.sub = FlowsheetBlock()

        assert m.fs.sub.flowsheet() is m.fs

    @pytest.mark.unit
    def test_default(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock()
        m.fs.sub = FlowsheetBlock()

        assert m.fs.sub.config.dynamic is False
        assert m.fs.sub.config.time is m.fs.config.time
        assert m.fs.sub.time_units is None

    @pytest.mark.unit
    def test_parent_dynamic_inherit(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        m.fs.sub = FlowsheetBlock()

        assert m.fs.sub.config.dynamic is True
        assert m.fs.sub.config.time is m.fs.config.time
        assert m.fs.sub.time_units == units.s

    @pytest.mark.unit
    def test_both_dynamic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        m.fs.sub = FlowsheetBlock(dynamic=True)

        assert m.fs.sub.config.dynamic is True
        assert m.fs.sub.config.time is m.fs.config.time
        assert m.fs.sub.time_units == units.s

    @pytest.mark.unit
    def test_ss_in_dynamic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        m.fs.sub = FlowsheetBlock(dynamic=False)

        assert m.fs.sub.config.dynamic is False
        assert m.fs.sub.config.time is m.fs.config.time
        assert m.fs.sub.time_units == units.s

    @pytest.mark.unit
    def test_dynamic_in_ss(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        with pytest.raises(DynamicError):
            m.fs.sub = FlowsheetBlock(dynamic=True)

    @pytest.mark.unit
    def test_ss_external_time(self):
        m = ConcreteModel()
        m.s = Set(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        m.fs.sub = FlowsheetBlock(dynamic=False, time=m.s)

        assert m.fs.sub.config.dynamic is False
        assert m.fs.sub.config.time is m.s
        assert m.fs.sub.time_units is None

    @pytest.mark.unit
    def test_dynamic_external_time(self):
        m = ConcreteModel()
        m.s = ContinuousSet(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        m.fs.sub = FlowsheetBlock(dynamic=True, time=m.s, time_units=units.minute)

        assert m.fs.sub.config.dynamic is True
        assert m.fs.sub.config.time is m.s
        assert m.fs.sub.time_units == units.minute

    @pytest.mark.unit
    def test_dynamic_external_time_invalid(self):
        m = ConcreteModel()
        m.s = Set(initialize=[4, 5])
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        with pytest.raises(DynamicError):
            m.fs.sub = FlowsheetBlock(dynamic=True, time=m.s, time_units=units.minute)

    @pytest.mark.unit
    def testtime_units_inherit(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_units=units.s)
        # Set different time units here to make sure they are ignored
        m.fs.sub = FlowsheetBlock(time_units=units.min)

        # Time should come from parent, not local settings
        assert m.fs.sub.time_units is units.s


class TestOther(object):
    @pytest.mark.unit
    def test_model_checks(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = PhysicalParameterTestBlock()
        m.fs.config.default_property_package = m.fs.props

        m.fs.unit1 = UnitModelBlock()

        m.fs.model_check()


class TestVisualisation(object):
    @pytest.mark.unit
    def test_report_empty(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.report()

    @pytest.mark.unit
    def test_get_stream_table_contents(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = PhysicalParameterTestBlock()
        m.fs.config.default_property_package = m.fs.props

        m.fs.unit1 = Heater()
        m.fs.unit2 = Heater()

        m.fs.stream = Arc(source=m.fs.unit1.outlet, destination=m.fs.unit2.inlet)
        TransformationFactory("network.expand_arcs").apply_to(m)

        df = m.fs._get_stream_table_contents()

        assert df.loc["pressure"]["stream"] == 1e5
        assert df.loc["temperature"]["stream"] == 300
        assert df.loc["component_flow_phase ('p1', 'c1')"]["stream"] == 2.0
        assert df.loc["component_flow_phase ('p1', 'c2')"]["stream"] == 2.0
        assert df.loc["component_flow_phase ('p2', 'c1')"]["stream"] == 2.0
        assert df.loc["component_flow_phase ('p2', 'c2')"]["stream"] == 2.0

        m.fs.report()


@pytest.mark.unit
# with this decorator, the RuntimeWarning will be raised as an error
@pytest.mark.filterwarnings("error::RuntimeWarning")
def test_ui_warnings():
    # force the UI to not seem to be installed
    ui = flowsheet_model.UI()
    ui.visualize = ui._visualize_null
    ui.installed = False
    # Call the function, which should print a warning
    try:
        ui.visualize(1, 2)
        assert False  # should not get here
    except RuntimeWarning:
        pass  # cool



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Tests for the controller model subclass of block
"""

import pytest
import pyomo.environ as pyo

from idaes.apps.caprese.tests.test_simple_model import (
    make_model,
    make_small_model,
    initialize_t0,
    copy_values_forward,
)
from idaes.apps.caprese.controller import (
    _ControllerBlockData,
    ControllerBlock,
    SimpleControllerBlock,
    IndexedControllerBlock,
)
from idaes.apps.caprese.nmpc_var import (
    DiffVar,
    AlgVar,
    InputVar,
    FixedVar,
    DerivVar,
)
from idaes.apps.caprese.common.config import (
    ControlPenaltyType,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

solver_available = pyo.SolverFactory("ipopt").available()
if solver_available:
    solver = get_solver(solver="ipopt")
else:
    solver = None


class TestControllerBlock(object):
    @pytest.mark.unit
    def test_construct(self):
        model = make_small_model()
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        controller = ControllerBlock(
            model=model,
            time=time,
            inputs=inputs,
        )
        assert type(controller) is SimpleControllerBlock
        assert isinstance(controller, ControllerBlock)
        assert isinstance(controller, _ControllerBlockData)

        controller.construct()
        assert controller[None] is controller
        assert controller.mod is model
        assert controller.time is time
        assert all(i1 is i2 for i1, i2 in zip(controller._inputs, inputs))

    @pytest.mark.unit
    def test_construct_indexed(self):
        controller_set = pyo.Set(initialize=[0, 1, 2])
        controller_set.construct()
        horizon_map = {0: 1.0, 1: 3.0, 2: 5.0}
        nfe_map = {0: 2, 1: 6, 2: 10}
        model_map = {i: make_model(horizon_map[i], nfe_map[i]) for i in controller_set}
        time_map = {i: model_map[i].time for i in controller_set}
        inputs_map = {i: [model_map[i].flow_in[0]] for i in controller_set}

        controller = ControllerBlock(
            controller_set,
            model=model_map,
            time=time_map,
            inputs=inputs_map,
        )

        assert type(controller) is IndexedControllerBlock
        assert isinstance(controller, IndexedControllerBlock)
        controller.construct()
        assert all(b.parent_component() is controller for b in controller.values())

        for i in controller_set:
            assert i in controller
        for i, c in controller.items():
            assert c.mod is model_map[i]
            assert c.time is time_map[i]
            assert all(i1 is i2 for i1, i2 in zip(c._inputs, inputs_map[i]))

    @pytest.mark.unit
    def make_controller(self, sample_time=0.5, horizon=1.0, nfe=2):
        model = make_model(horizon=horizon, nfe=nfe)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        controller = ControllerBlock(
            model=model,
            time=time,
            inputs=inputs,
        )
        controller.construct()
        controller.set_sample_time(sample_time)
        return controller

    @pytest.mark.unit
    def test_add_setpoint_objective(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
            (controller.mod.flow_in[t0], 3.0),
        ]
        weights = [
            (controller.mod.flow_in[t0], 2.0),
        ]
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)

        controller.add_setpoint_objective(setpoint, weights)

        assert hasattr(controller, "setpoint_objective")

        pred_obj_expr = 2.0 * (controller.mod.flow_in[t0] - 3.0) ** 2
        obj_expr = controller.setpoint_objective.expr
        assert pyo.value(pred_obj_expr) == pyo.value(obj_expr)
        assert pred_obj_expr.to_string() == obj_expr.to_string()

        controller.del_component(controller.setpoint_objective)

        setpoint = [
            (controller.mod.flow_in[t0], 3.0),
            (controller.mod.conc[t0, "A"], 1.5),
        ]
        weights = [
            (controller.mod.flow_in[t0], 1.0),
            (controller.mod.conc[t0, "A"], 5.0),
        ]
        controller.add_setpoint_objective(setpoint, weights)
        pred_obj_expr = (
            1.0 * (controller.mod.flow_in[t0] - 3.0) ** 2
            + 5.0 * (controller.mod.conc[t0, "A"] - 1.5) ** 2
        )
        obj_expr = controller.setpoint_objective.expr
        assert pyo.value(pred_obj_expr) == pyo.value(obj_expr)
        assert pred_obj_expr.to_string() == obj_expr.to_string()

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason="IPOPT is not available")
    def test_solve_setpoint_steady(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
            (controller.mod.flow_in[t0], 3.0),
        ]
        weights = [
            (controller.mod.flow_in[t0], 1.0),
        ]
        controller.add_setpoint_objective(setpoint, weights)
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)

        dof_prior = degrees_of_freedom(controller)
        controller.solve_setpoint(solver, require_steady=True)
        dof_post = degrees_of_freedom(controller)

        assert dof_prior == dof_post

        assert controller.differential_vars[0].setpoint == pytest.approx(3.75, abs=1e-3)
        assert controller.differential_vars[1].setpoint == pytest.approx(1.25, abs=1e-3)
        assert controller.algebraic_vars[0].setpoint == pytest.approx(3.0, abs=1e-3)
        assert controller.algebraic_vars[1].setpoint == pytest.approx(-3.75, abs=1e-3)
        assert controller.algebraic_vars[2].setpoint == pytest.approx(3.75, abs=1e-3)
        assert controller.input_vars[0].setpoint == pytest.approx(3.0, abs=1e-3)

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason="IPOPT is not available")
    def test_solve_setpoint_unsteady(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
            (controller.mod.flow_in[t0], 3.0),
            (controller.mod.conc[t0, "A"], 5.0),
        ]
        weights = [
            (controller.mod.flow_in[t0], 1.0),
            (controller.mod.conc[t0, "A"], 1.0),
        ]
        controller.add_setpoint_objective(setpoint, weights)
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)

        dof_prior = degrees_of_freedom(controller)
        controller.solve_setpoint(solver, require_steady=False)
        dof_post = degrees_of_freedom(controller)

        assert dof_prior == dof_post

        assert controller.differential_vars[0].setpoint == pytest.approx(5.0, abs=1e-3)
        assert controller.differential_vars[1].setpoint == pytest.approx(7.11, abs=1e-3)
        assert controller.algebraic_vars[0].setpoint == pytest.approx(3.0, abs=1e-3)
        assert controller.algebraic_vars[1].setpoint == pytest.approx(-5.0, abs=1e-3)
        assert controller.algebraic_vars[2].setpoint == pytest.approx(5.0, abs=1e-3)
        assert controller.input_vars[0].setpoint == pytest.approx(3.0, abs=1e-3)

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason="IPOPT is not available")
    def test_add_tracking_objective(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
            (controller.mod.flow_in[t0], 3.0),
        ]
        weights = [
            (controller.mod.flow_in[t0], 1.0),
        ]
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)
        copy_values_forward(controller.mod)
        controller.add_setpoint_objective(setpoint, weights)
        controller.solve_setpoint(solver)

        # Re-initialize inputs so they are not at the setpoint
        # for objective evaluation.
        controller.mod.flow_in[:].set_value(2.5)

        weights = [
            (controller.mod.conc[t0, "A"], 1),
            (controller.mod.conc[t0, "B"], 1),
            (controller.mod.rate[t0, "A"], 1),
            (controller.mod.rate[t0, "B"], 1),
            (controller.mod.flow_out[t0], 1),
            (controller.mod.flow_in[t0], 1),
        ]

        # Construct predicted objective functions:
        pred_obj = {i: 0.0 for i in range(1, 4)}
        sample_points = controller.sample_points[1:]
        sample_time = controller.sample_time
        for v in controller.component_objects(DiffVar):
            for t in sample_points:
                pred_obj[1] += (v[t] - v.setpoint) ** 2
                pred_obj[2] += (v[t] - v.setpoint) ** 2
                pred_obj[3] += (v[t] - v.setpoint) ** 2
        for v in controller.component_objects(AlgVar):
            for t in sample_points:
                pred_obj[3] += (v[t] - v.setpoint) ** 2
        for v in controller.component_objects(InputVar):
            for t in sample_points:
                i_prev = time.find_nearest_index(t - sample_time, tolerance=1e-8)
                t_prev = time.at(i_prev)
                pred_obj[1] += (v[t] - v.setpoint) ** 2
                pred_obj[2] += (v[t] - v[t_prev]) ** 2
                pred_obj[3] += (v[t] - v[t_prev]) ** 2

        controller.add_tracking_objective(
            weights,
            control_penalty_type=ControlPenaltyType.ERROR,
            state_ctypes=DiffVar,
        )
        assert pyo.value(controller.tracking_objective.expr == pred_obj[1])
        assert pyo.value(controller.tracking_objective.expr) > 0
        controller.del_component(controller.tracking_objective)

        controller.add_tracking_objective(
            weights,
            control_penalty_type=ControlPenaltyType.ACTION,
            state_ctypes=DiffVar,
        )
        assert pyo.value(controller.tracking_objective.expr == pred_obj[2])
        assert pyo.value(controller.tracking_objective.expr) > 0
        controller.del_component(controller.tracking_objective)

        controller.add_tracking_objective(
            weights,
            control_penalty_type=ControlPenaltyType.ACTION,
            state_ctypes=(DiffVar, AlgVar),
        )
        assert pyo.value(controller.tracking_objective.expr == pred_obj[3])
        assert pyo.value(controller.tracking_objective.expr) > 0

    @pytest.mark.unit
    def test_constrain_control_inputs_piecewise_constant(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()

        controller.constrain_control_inputs_piecewise_constant()
        assert hasattr(controller, "pwc_constraint")
        sample_points = controller.sample_points
        sample_point_set = set(sample_points)

        for i, t in controller.INPUT_SET * controller.time:
            if t in sample_point_set:
                assert (i, t) not in controller.pwc_constraint
            else:
                assert (i, t) in controller.pwc_constraint
        for i, t in controller.pwc_constraint:
            pwc_expr = controller.pwc_constraint[i, t].expr
            tn = time.next(t)
            inputs = controller.vectors.input
            pred_expr = inputs[i, tn] == inputs[i, t]
            assert pwc_expr.to_string() == pred_expr.to_string()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""Basic unit tests for PETSc solver utilities"""
import pytest
import re
import numpy as np
import json
import os
import pyomo.environ as pyo
import pyomo.dae as pyodae
from pyomo.util.subsystems import create_subsystem_block
from idaes.core.solvers import petsc
import idaes.logger as idaeslog


def rp_example():
    """This example is done multiple ways to test a few common errors where the
    integrator and fully time-discretized problem differ and alternative working
    formulations.

    The PETSc utilities raise an exception when a differential variable is fixed.
    While this is okay for the fully time-discretized problem, the integrator
    will not correctly link a fixed differential variable (at non-initial time
    points) with a time derivative.
    """
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=1, scheme="BACKWARD")

    for t in m.time:
        m.x[t].fix(2.0 * t)

    m.u[0].fix(1.0)

    return m


def rp_example2():
    """This example is done multiple ways to test a few common errors where the
    integrator and fully time-discretized problem differ, and alternative working
    formulations.

    Rather than fixing a differential variable, we can add an explicit time
    variable for the integrator and fix the time variable to the time index
    for the fully discretized problem. This works as an alternative to the
    fixed differential variable.
    """
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.t = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    def x_eq_rule(m, t):
        return m.x[t] == 2.0 * m.t[t]

    m.x_eq = pyo.Constraint(m.time, rule=x_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=1, scheme="BACKWARD")

    m.u[0].fix(1.0)
    # For fully discretized fix all times at time index
    m.t[0].fix(m.time.first())

    return m


def rp_example3():
    """This example is done multiple ways to test a few common errors where the
    integrator and fully time-discretized problem differ, and alternative working
    formulations.

    Another way to formulate this problem is to fix the derivative. This doesn't
    work for the integrator because it loses the association between the
    differential variable and its derivative.  Since for users, the result of
    this formulation may be unexpected, the PETSc utilities will raise an
    exception if derivatives are fixed to anything other than 0.
    """
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=1, scheme="BACKWARD")

    m.u[0].fix(1.0)
    m.dxdt[:].fix(2.0)

    return m


def rp_example4():
    """This example is done multiple ways to test a few common errors where the
    integrator and fully time-discretized problem differ, and alternative working
    formulations.

    Rather than fixing the derivative, we can add a constraint to set the
    derivative.  This should work as intended for both the fully
    time-discretized problem and integrator.
    """
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time, initialize=1)
    m.u = pyo.Var(m.time, initialize=1)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq1_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq1 = pyo.Constraint(m.time, rule=diff_eq1_rule)

    def diff_eq2_rule(m, t):
        return m.dxdt[t] == 2.0

    m.diff_eq2 = pyo.Constraint(m.time, rule=diff_eq2_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=1, scheme="BACKWARD")

    m.u[0].fix(1.0)
    m.x[0].fix(0.0)
    m.diff_eq2[0].deactivate()

    return m


def rp_example5():
    """This example is used to test ramping only in a subset of time."""
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.ramp = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    def diff_eq2_rule(m, t):
        return m.dxdt[t] == m.ramp[t]

    m.diff_eq2 = pyo.Constraint(m.time, rule=diff_eq2_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=10, scheme="BACKWARD")

    m.x[0].fix(0.0)
    for t, v in m.ramp.items():
        if t > 3 and t <= 5:
            v.fix(4)
        else:
            v.fix(0)

    assert pyo.value(m.ramp[0]) == 0
    assert pyo.value(m.ramp[1]) == 0
    assert pyo.value(m.ramp[2]) == 0
    assert pyo.value(m.ramp[3]) == 0
    assert pyo.value(m.ramp[4]) == 4
    assert pyo.value(m.ramp[5]) == 4
    assert pyo.value(m.ramp[6]) == 0
    assert pyo.value(m.ramp[7]) == 0
    assert pyo.value(m.ramp[8]) == 0
    assert pyo.value(m.ramp[9]) == 0
    assert pyo.value(m.ramp[10]) == 0

    return m


def car_example():
    """This is to test problems where a differential variable doesn't appear in
    a constraint this is based on a Pyomo example here:
    https://github.com/Pyomo/pyomo/blob/main/examples/dae/car_example.py"""
    m = pyo.ConcreteModel()

    m.R = pyo.Param(initialize=0.001)  #  Friction factor
    m.L = pyo.Param(initialize=100.0)  #  Final position

    m.tau = pyodae.ContinuousSet(bounds=(0, 1))  # Unscaled time
    m.time = pyo.Var(m.tau)  # Scaled time
    m.tf = pyo.Var()
    m.x = pyo.Var(m.tau, bounds=(0, m.L + 50))
    m.v = pyo.Var(m.tau, bounds=(0, None))
    m.a = pyo.Var(m.tau, bounds=(-3.0, 1.0), initialize=0)

    m.dtime = pyodae.DerivativeVar(m.time)
    m.dx = pyodae.DerivativeVar(m.x)
    m.dv = pyodae.DerivativeVar(m.v)

    m.obj = pyo.Objective(expr=m.tf)

    def _ode1(m, i):
        if i == 0:
            return pyo.Constraint.Skip
        return m.dx[i] == m.tf * m.v[i]

    m.ode1 = pyo.Constraint(m.tau, rule=_ode1)

    def _ode2(m, i):
        if i == 0:
            return pyo.Constraint.Skip
        return m.dv[i] == m.tf * (m.a[i] - m.R * m.v[i] ** 2)

    m.ode2 = pyo.Constraint(m.tau, rule=_ode2)

    def _ode3(m, i):
        if i == 0:
            return pyo.Constraint.Skip
        return m.dtime[i] == m.tf

    m.ode3 = pyo.Constraint(m.tau, rule=_ode3)

    def _init(m):
        yield m.x[0] == 0
        # yield m.x[1] == m.L
        yield m.v[0] == 0
        yield m.v[1] == 0
        yield m.time[0] == 0

    m.initcon = pyo.ConstraintList(rule=_init)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=1, scheme="BACKWARD")
    return m


def dae_with_non_time_indexed_constraint(
    nfe=1, ncp=3, transformation_method="dae.finite_difference", scheme="BACKWARD"
):
    """This function provides a DAE model for solver testing. This model contains a non-
    time-indexed variable and constraint and a fixed derivative to test some
    edge cases.

    The problem and expected result are from A test problem from
    https://archimede.dm.uniba.it/~testset/report/chemakzo.pdf.

    Args:
        nfe: Number of finite elements to use in discretization
        ncp: Number of collocation points to use, if using collocation
        transformation_method: Discretization method. Presently,
            options are "dae.finite_difference" and "dae.collocation".
        scheme: Scheme to use in discretization method. Check Pyomo
            DAE documentation for more info

    Returns:
        (tuple): Pyomo ConcreteModel, correct solved value for y[1] to y[6]
    """
    model = pyo.ConcreteModel(name="chemakzo")

    # Set problem parameter values
    model.k = pyo.Param([1, 2, 3, 4], initialize={1: 18.7, 2: 0.58, 3: 0.09, 4: 0.42})
    model.Ke = pyo.Param(initialize=34.4)
    model.klA = pyo.Param(initialize=3.3)
    model.Ks = pyo.Param(initialize=115.83)
    model.pCO2 = pyo.Param(initialize=0.9)
    # The following parameter H, is best made a parameter, but will use a
    # variable and constraint instead to test non-time-indexed constraints
    # model.H = pyo.Param(initialize=737)

    # Problem variables ydot = dy/dt,
    #    (dy6/dt is not explicitly in the equations, so only 5 ydots)
    model.H = pyo.Var(initialize=737)
    model.t = pyodae.ContinuousSet(bounds=(0, 180))
    model.y = pyo.Var(model.t, [1, 2, 3, 4, 5, 6], initialize=1.0)  #
    model.ydot = pyodae.DerivativeVar(model.y, wrt=model.t)  # dy/dt
    model.r = pyo.Var(model.t, [1, 2, 3, 4, 5], initialize=1.0)
    model.Fin = pyo.Var(model.t, initialize=1.0)

    # Non-time indexed constraint (just for testing)
    model.H_eqn = pyo.Constraint(expr=model.H == 737)

    # Equations
    @model.Constraint(model.t)
    def eq_ydot1(b, t):
        return b.ydot[t, 1] == -2.0 * b.r[t, 1] + b.r[t, 2] - b.r[t, 3] - b.r[t, 4]

    @model.Constraint(model.t)
    def eq_ydot2(b, t):
        return b.ydot[t, 2] == -0.5 * b.r[t, 1] - b.r[t, 4] - 0.5 * b.r[t, 5] + b.Fin[t]

    @model.Constraint(model.t)
    def eq_ydot3(b, t):
        return b.ydot[t, 3] == b.r[t, 1] - b.r[t, 2] + b.r[t, 3]

    @model.Constraint(model.t)
    def eq_ydot4(b, t):
        return b.ydot[t, 4] == -b.r[t, 2] + b.r[t, 3] - 2.0 * b.r[t, 4]

    @model.Constraint(model.t)
    def eq_ydot5(b, t):
        return b.ydot[t, 5] == b.r[t, 2] - b.r[t, 3] + b.r[t, 5]

    @model.Constraint(model.t)
    def eq_y6(b, t):
        return 0 == b.Ks * b.y[t, 1] * b.y[t, 4] - b.y[t, 6]

    @model.Constraint(model.t)
    def eq_r1(b, t):
        return b.r[t, 1] == b.k[1] * b.y[t, 1] ** 4 * b.y[t, 2] ** 0.5

    @model.Constraint(model.t)
    def eq_r2(b, t):
        return b.r[t, 2] == b.k[2] * b.y[t, 3] * b.y[t, 4]

    @model.Constraint(model.t)
    def eq_r3(b, t):
        return b.r[t, 3] == b.k[2] / b.Ke * b.y[t, 1] * b.y[t, 5]

    @model.Constraint(model.t)
    def eq_r4(b, t):
        return b.r[t, 4] == b.k[3] * b.y[t, 1] * b.y[t, 4] ** 2

    @model.Constraint(model.t)
    def eq_r5(b, t):
        return b.r[t, 5] == b.k[4] * b.y[t, 6] ** 2 * b.y[t, 2] ** 0.5

    @model.Constraint(model.t)
    def eq_Fin(b, t):
        return b.Fin[t] == b.klA * (b.pCO2 / b.H - b.y[t, 2])

    # Set initial conditions and solve initial from the values of differential
    # variables (r and y6 well and the derivative vars too).
    y0 = {1: 0.444, 2: 0.00123, 3: 0.0, 4: 0.007, 5: 0.0}  # initial differential vars
    for i in y0:
        model.y[0, i].fix(y0[i])

    if scheme == "BACKWARD":
        model.eq_ydot1[0].deactivate()
        model.eq_ydot2[0].deactivate()
        model.eq_ydot3[0].deactivate()
        model.eq_ydot4[0].deactivate()
        model.eq_ydot5[0].deactivate()

    discretizer = pyo.TransformationFactory(transformation_method)
    if transformation_method == "dae.collocation":
        discretizer.apply_to(model, nfe=nfe, ncp=ncp, scheme=scheme)
    else:
        discretizer.apply_to(model, nfe=nfe, scheme=scheme)
    model.ydot[:, 6].fix(0)

    return (
        model,
        0.1150794920661702,
        0.1203831471567715e-2,
        0.1611562887407974,
        0.3656156421249283e-3,
        0.1708010885264404e-1,
        0.4873531310307455e-2,
    )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_car():
    m = car_example()
    m.a.fix(1.0)
    m.tf.fix(16.56)

    # solve
    petsc.petsc_dae_by_time_element(
        m,
        time=m.tau,
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
        },
    )

    assert pyo.value(m.x[1]) == pytest.approx(131.273, rel=1e-2)
    petsc.calculate_time_derivatives(m, m.tau)
    assert pyo.value(m.dtime[1]) == pytest.approx(pyo.value(m.time[1] - m.time[0]))
    assert pyo.value(m.dx[1]) == pytest.approx(pyo.value(m.x[1] - m.x[0]))
    assert pyo.value(m.dv[1]) == pytest.approx(pyo.value(m.v[1] - m.v[0]))


@pytest.mark.unit
def test_copy_time():
    """test the time copy function.  When this is used, the model is flattened
    and only indexed by time, so testing is pretty simple"""

    m = pyo.ConcreteModel()
    t = [1, 2]
    m.x = pyo.Var(t)
    m.y = pyo.Var(t)

    m.x[1] = 1
    m.x[2] = 2
    m.y[1] = 3
    m.y[2] = 4

    petsc._copy_time([m.x, m.y], 1, 2)

    assert pyo.value(m.x[2]) == 1
    assert pyo.value(m.y[2]) == 3


@pytest.mark.unit
def test_gen_time_disc_eqns():
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint()

    # The model has one time element and derivatives for y[1] to y[5]
    # the final time is 180, so create a list of what the time discretization
    # constraints should be.
    disc_eq = [
        id(m.ydot_disc_eq[180, 1]),
        id(m.ydot_disc_eq[180, 2]),
        id(m.ydot_disc_eq[180, 3]),
        id(m.ydot_disc_eq[180, 4]),
        id(m.ydot_disc_eq[180, 5]),
    ]

    n = 0
    for cs in petsc.find_discretization_equations(m, m.t):
        for c in cs.values():
            if c.index()[1] == 6:
                continue
            assert id(c) in disc_eq
            n += 1

    assert len(disc_eq) == n


@pytest.mark.unit
def test_set_dae_suffix():
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint()
    regular_vars, time_vars = pyodae.flatten.flatten_dae_components(m, m.t, pyo.Var)
    regular_cons, time_cons = pyodae.flatten.flatten_dae_components(
        m, m.t, pyo.Constraint
    )
    t = 180
    m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
    m.scaling_factor[m.ydot[180, 2]] = 10
    constraints = [con[t] for con in time_cons if t in con]
    variables = [var[t] for var in time_vars]
    t_block = create_subsystem_block(constraints, variables)
    deriv_diff_map = petsc._get_derivative_differential_data_map(m, m.t)
    petsc._set_dae_suffixes_from_variables(t_block, variables, deriv_diff_map)
    petsc._sub_problem_scaling_suffix(m, t_block)

    assert t_block.dae_suffix[m.ydot[180, 1]] == 2
    assert t_block.dae_suffix[m.ydot[180, 2]] == 2
    assert t_block.dae_suffix[m.ydot[180, 3]] == 2
    assert t_block.dae_suffix[m.ydot[180, 4]] == 2
    assert t_block.dae_suffix[m.ydot[180, 5]] == 2
    assert t_block.scaling_factor[m.ydot[180, 2]] == 10
    assert t_block.dae_suffix[m.y[180, 1]] == 1
    assert t_block.dae_suffix[m.y[180, 2]] == 1
    assert t_block.dae_suffix[m.y[180, 3]] == 1
    assert t_block.dae_suffix[m.y[180, 4]] == 1
    assert t_block.dae_suffix[m.y[180, 5]] == 1

    # Make sure deactivating a differential equation makes the variable that
    # would have been differential go algebraic
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint()
    # discretization equations would be deactivated in normal PETSc solve
    for con in petsc.find_discretization_equations(m, m.t):
        con.deactivate()
    # deactivate a differential equation making y4 be algebraic
    m.eq_ydot4[180].deactivate()
    regular_vars, time_vars = pyodae.flatten.flatten_dae_components(m, m.t, pyo.Var)
    regular_cons, time_cons = pyodae.flatten.flatten_dae_components(
        m, m.t, pyo.Constraint
    )
    t = 180
    constraints = [con[t] for con in time_cons if t in con]
    variables = [var[t] for var in time_vars]
    t_block = create_subsystem_block(constraints, variables)
    deriv_diff_map = petsc._get_derivative_differential_data_map(m, m.t)
    petsc._set_dae_suffixes_from_variables(t_block, variables, deriv_diff_map)
    petsc._sub_problem_scaling_suffix(m, t_block)

    assert m.y[t, 4] not in t_block.dae_suffix
    assert m.ydot[t, 4] not in t_block.dae_suffix


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_petsc_read_trajectory():
    """
    Check that the PETSc DAE solver works.
    """
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint()
    m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
    m.scaling_factor[m.y[180, 1]] = 10  # make sure unscale works

    m.y_ref = pyo.Reference(m.y)  # make sure references don't get unscaled twice
    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.t,
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
            "--ts_save_trajectory": 1,
            "--ts_trajectory_type": "visualization",
        },
    )
    assert pytest.approx(y1, rel=1e-3) == pyo.value(m.y[m.t.last(), 1])
    assert pytest.approx(y2, rel=1e-3) == pyo.value(m.y[m.t.last(), 2])
    assert pytest.approx(y3, rel=1e-3) == pyo.value(m.y[m.t.last(), 3])
    assert pytest.approx(y4, rel=1e-3) == pyo.value(m.y[m.t.last(), 4])
    assert pytest.approx(y5, rel=1e-3) == pyo.value(m.y[m.t.last(), 5])
    assert pytest.approx(y6, rel=1e-3) == pyo.value(m.y[m.t.last(), 6])

    tj = res.trajectory
    assert tj.get_dt()[0] == pytest.approx(0.01)  # if small enough shouldn't be cut
    assert tj.get_vec(m.y[180, 1])[-1] == pytest.approx(y1, rel=1e-3)
    assert tj.get_vec("_time")[-1] == pytest.approx(180)

    times = np.linspace(0, 180, 181)
    tj2 = tj.interpolate(times)
    assert tj2.get_vec(m.y[180, 1])[180] == pytest.approx(y1, rel=1e-3)
    assert tj2.time[180] == pytest.approx(180)

    tj.to_json("some_testy_json.json")
    with open("some_testy_json.json", "r") as fp:
        vecs = json.load(fp)
    assert vecs[str(m.y[180, 1])][-1] == pytest.approx(y1, rel=1e-3)
    assert vecs["_time"][-1] == pytest.approx(180)
    os.remove("some_testy_json.json")

    tj.to_json("some_testy_json.json.gz")
    tj2 = petsc.PetscTrajectory(json="some_testy_json.json.gz")
    assert tj2.vecs[str(m.y[180, 1])][-1] == pytest.approx(y1, rel=1e-3)
    assert tj2.vecs["_time"][-1] == pytest.approx(180)
    os.remove("some_testy_json.json.gz")

    tj2 = petsc.PetscTrajectory(vecs=vecs)
    assert tj2.vecs[str(m.y[180, 1])][-1] == pytest.approx(y1, rel=1e-3)
    assert tj2.vecs["_time"][-1] == pytest.approx(180)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_petsc_read_trajectory_parts():
    """
    Check that the PETSc DAE solver works.
    """
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint(nfe=10)
    m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
    m.scaling_factor[m.y[180, 1]] = 10  # make sure unscale works

    m.y_ref = pyo.Reference(m.y)  # make sure references don't get unscaled twice
    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.t,
        between=[m.t.first(), m.t.at(4), m.t.last()],
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
            "--ts_save_trajectory": 1,
        },
    )
    assert pytest.approx(y1, rel=1e-3) == pyo.value(m.y[m.t.last(), 1])
    assert pytest.approx(y2, rel=1e-3) == pyo.value(m.y[m.t.last(), 2])
    assert pytest.approx(y3, rel=1e-3) == pyo.value(m.y[m.t.last(), 3])
    assert pytest.approx(y4, rel=1e-3) == pyo.value(m.y[m.t.last(), 4])
    assert pytest.approx(y5, rel=1e-3) == pyo.value(m.y[m.t.last(), 5])
    assert pytest.approx(y6, rel=1e-3) == pyo.value(m.y[m.t.last(), 6])

    tj = res.trajectory
    assert tj.get_vec(m.y[180, 1])[-1] == pytest.approx(y1, rel=1e-3)
    assert tj.get_vec("_time")[-1] == pytest.approx(180)
    y1_trj = tj.interpolate_vec(m.t, m.y[180, 1])
    y4_trj = tj.interpolate_vec(m.t, m.y[180, 4])
    for i, t in enumerate(m.t):
        assert y1_trj[i] == pytest.approx(pyo.value(m.y[t, 1]))
        assert y4_trj[i] == pytest.approx(pyo.value(m.y[t, 4]))


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_rp_example():

    m = rp_example()
    with pytest.raises(RuntimeError):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.time,
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_rp_example2():

    m = rp_example2()
    petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        timevar=m.t,
        ts_options={
            "--ts_dt": 1,
            "--ts_adapt_type": "none",
        },
    )
    assert pyo.value(m.u[10]) == pytest.approx(398)
    assert pyo.value(m.x[10]) == pytest.approx(20)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_rp_example3():

    m = rp_example3()
    with pytest.raises(RuntimeError):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.time,
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_rp_example4():

    m = rp_example4()
    petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        ts_options={
            "--ts_dt": 1,
            "--ts_adapt_type": "none",
        },
    )
    assert pyo.value(m.u[10]) == pytest.approx(398)
    assert pyo.value(m.x[10]) == pytest.approx(20)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_rp_example5a():

    m = rp_example5()
    petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[0, 3, 5, 10],
        ts_options={
            "--ts_dt": 1,
            "--ts_adapt_type": "none",
            "--ts_save_trajectory": 1,
        },
    )

    assert pyo.value(m.x[0]) == pytest.approx(0)
    assert pyo.value(m.x[1]) == pytest.approx(0)
    assert pyo.value(m.x[2]) == pytest.approx(0)
    assert pyo.value(m.x[3]) == pytest.approx(0)
    assert pyo.value(m.x[4]) == pytest.approx(4)
    assert pyo.value(m.x[5]) == pytest.approx(8)
    assert pyo.value(m.x[6]) == pytest.approx(8)
    assert pyo.value(m.x[7]) == pytest.approx(8)
    assert pyo.value(m.x[8]) == pytest.approx(8)
    assert pyo.value(m.x[9]) == pytest.approx(8)
    assert pyo.value(m.x[10]) == pytest.approx(8)

    assert pyo.value(m.u[0]) == pytest.approx(0)
    assert pyo.value(m.u[1]) == pytest.approx(0)
    assert pyo.value(m.u[2]) == pytest.approx(0)
    assert pyo.value(m.u[3]) == pytest.approx(0)
    assert pyo.value(m.u[4]) == pytest.approx(12)
    assert pyo.value(m.u[5]) == pytest.approx(60)
    assert pyo.value(m.u[6]) == pytest.approx(64)
    assert pyo.value(m.u[7]) == pytest.approx(64)
    assert pyo.value(m.u[8]) == pytest.approx(64)
    assert pyo.value(m.u[9]) == pytest.approx(64)
    assert pyo.value(m.u[10]) == pytest.approx(64)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_rp_example5b():
    m = rp_example5()
    with pytest.raises(RuntimeError):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.time,
            between=[0, 3.5, 5, 10],
            ts_options={
                "--ts_dt": 1,
                "--ts_adapt_type": "none",
                "--ts_save_trajectory": 1,
            },
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_mixed_derivative_exception():
    # Test exception when mixed space/time derivatives appear in problem
    m = pyo.ConcreteModel()

    m.R = pyo.Param(initialize=0.001)  #  Friction factor
    m.L = pyo.Param(initialize=100.0)  #  Final position

    m.t = pyodae.ContinuousSet(bounds=(0, 1))
    m.x = pyodae.ContinuousSet(bounds=(0, 1))

    m.temperature = pyo.Var(m.t, m.x)
    m.d2T_dtdx = pyodae.DerivativeVar(m.temperature, wrt=(m.t, m.x))

    # This problem isn't well-posed, but that shouldn't matter
    @m.Constraint(m.t, m.x)
    def constraint_eqn(b, t, x):
        return b.d2T_dtdx[t, x] == b.temperature[t, x] + 1

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=3, scheme="BACKWARD", wrt=m.t)
    discretizer.apply_to(m, nfe=3, scheme="BACKWARD", wrt=m.x)

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "IDAES presently does not support PETSc for second order or higher derivatives like d2T_dtdx "
            "that are differentiated at least once with respect to time. Please reformulate your model so "
            "it does not contain such a derivative (such as by introducing intermediate variables)."
        ),
    ):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.t,
            between=[0, 1],
            ts_options={
                "--ts_dt": 1,
                "--ts_adapt_type": "none",
                "--ts_save_trajectory": 1,
            },
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_petsc_skip_initial_solve():
    """
    Ensure skipping the initial solution works as intended
    """
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint()

    m.y[0, 6].value = 0.35999964

    m.r[0, 1].value = 0.025487429806082887
    m.r[0, 2].value = 0
    m.r[0, 3].value = 0
    m.r[0, 4].value = 1.9580400000000004e-06
    m.r[0, 5].value = 0.0019090002227229192

    m.H.value = 737
    m.Fin[0].value = -2.9149253731343243e-05

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.t,
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
            "--ts_save_trajectory": 1,
            "--ts_trajectory_type": "visualization",
        },
        skip_initial=True,
    )

    assert pytest.approx(y1, rel=1e-3) == pyo.value(m.y[m.t.last(), 1])
    assert pytest.approx(y2, rel=1e-3) == pyo.value(m.y[m.t.last(), 2])
    assert pytest.approx(y3, rel=1e-3) == pyo.value(m.y[m.t.last(), 3])
    assert pytest.approx(y4, rel=1e-3) == pyo.value(m.y[m.t.last(), 4])
    assert pytest.approx(y5, rel=1e-3) == pyo.value(m.y[m.t.last(), 5])
    assert pytest.approx(y6, rel=1e-3) == pyo.value(m.y[m.t.last(), 6])


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_petsc_traj_previous():
    """
    Make sure trajectory concatenation works as intended
    """
    m, y1, y2, y3, y4, y5, y6 = dae_with_non_time_indexed_constraint(nfe=10)

    # First get a reference trajectory from a single solve
    res0 = petsc.petsc_dae_by_time_element(
        m,
        time=m.t,
        between=[m.t.first(), m.t.last()],
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
            "--ts_save_trajectory": 1,
        },
    )
    tj0 = res0.trajectory

    # Next do it in two steps
    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.t,
        between=[m.t.first(), m.t.at(4)],
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
            "--ts_save_trajectory": 1,
        },
    )
    # Fix initial condition for second leg of trajectory
    for j in range(1, 6):
        m.y[m.t.at(4), j].fix()
    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.t,
        between=[m.t.at(4), m.t.last()],
        ts_options={
            "--ts_type": "cn",  # Crank–Nicolson
            "--ts_adapt_type": "basic",
            "--ts_dt": 0.01,
            "--ts_save_trajectory": 1,
        },
        previous_trajectory=res.trajectory,
        # interpolate=False
    )

    assert pytest.approx(y1, rel=1e-3) == pyo.value(m.y[m.t.last(), 1])
    assert pytest.approx(y2, rel=1e-3) == pyo.value(m.y[m.t.last(), 2])
    assert pytest.approx(y3, rel=1e-3) == pyo.value(m.y[m.t.last(), 3])
    assert pytest.approx(y4, rel=1e-3) == pyo.value(m.y[m.t.last(), 4])
    assert pytest.approx(y5, rel=1e-3) == pyo.value(m.y[m.t.last(), 5])
    assert pytest.approx(y6, rel=1e-3) == pyo.value(m.y[m.t.last(), 6])

    tj = res.trajectory
    assert tj.get_vec(m.y[180, 1])[-1] == pytest.approx(y1, rel=1e-3)
    assert tj.get_vec("_time")[-1] == pytest.approx(180)
    y1_trj = tj.interpolate_vec(m.t, m.y[180, 1])
    y4_trj = tj.interpolate_vec(m.t, m.y[180, 4])
    for i, t in enumerate(m.t):
        assert y1_trj[i] == pytest.approx(pyo.value(m.y[t, 1]))
        assert y4_trj[i] == pytest.approx(pyo.value(m.y[t, 4]))

    t_vec = np.linspace(0, 180, 101)
    y2_tj0 = tj0.interpolate_vec(t_vec, m.y[180, 2])
    y5_tj0 = tj0.interpolate_vec(t_vec, m.y[180, 5])
    y2_tj = tj.interpolate_vec(t_vec, m.y[180, 2])
    y5_tj = tj.interpolate_vec(t_vec, m.y[180, 5])

    for i, t in enumerate(t_vec):
        assert y2_tj[i] == pytest.approx(y2_tj0[i], abs=1e-4)
        assert y5_tj[i] == pytest.approx(y5_tj0[i], abs=1e-4)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_snes_options_deprecation(caplog):
    m = rp_example2()
    caplog.set_level(idaeslog.WARNING)
    petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        timevar=m.t,
        ts_options={
            "--ts_dt": 1,
            "--ts_adapt_type": "none",
            "--ts_monitor": "",
        },
        snes_options={},
    )
    s = ""
    for record in caplog.records:
        s += record.message
    assert "DEPRECATED" in s


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_double_options_exception():

    m = rp_example2()
    with pytest.raises(RuntimeError, match="deprecated"):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.time,
            timevar=m.t,
            snes_options={
                "--dummy": "",
            },
            initial_solver_options={
                "--dummy": "",
            },
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_not_ContinuousSet():
    m = pyo.ConcreteModel()

    m.time = pyo.Set(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)

    with pytest.raises(RuntimeError, match="Pyomo"):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.time,
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_not_discretized():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    with pytest.raises(RuntimeError, match="discretized"):
        petsc.petsc_dae_by_time_element(
            m,
            time=m.time,
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_representative_time_not_in_between():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 5.0, 10.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=2, scheme="BACKWARD")

    with pytest.raises(RuntimeError, match="representative_time"):
        petsc.petsc_dae_by_time_element(
            m, time=m.time, between=[0.0, 10.0], representative_time=5.0
        )


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_calculate_derivatives():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=2, scheme="BACKWARD")

    m.u.fix(1.0)
    m.x[0].fix(0.0)
    m.diff_eq[0].deactivate()

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[0.0, 2.0],
        ts_options={
            "--ts_type": "beuler",
            "--ts_dt": 3e-2,
        },
        skip_initial=True,  # With u and x fixed, no variables to solve for at t0
        calculate_derivatives=True,
    )
    # No value assigned at 0 because there's no corresponding discretization equation
    assert m.dxdt[0].value is None
    # It should backfill values for interpolated points like t=1
    assert pyo.value(m.dxdt[1]) == pytest.approx(-0.7580125427537873, rel=1e-3)
    assert pyo.value(m.dxdt[2]) == pytest.approx(-0.2034733131369807, rel=1e-3)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_calculate_derivatives_integrate_first_half():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=2, scheme="BACKWARD")

    m.u.fix(1.0)
    m.x[0].fix(0.0)
    m.diff_eq[0].deactivate()

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[0.0, 1.0],
        ts_options={
            "--ts_type": "beuler",
            "--ts_dt": 3e-2,
        },
        skip_initial=True,  # With u and x fixed, no variables to solve for at t0
        calculate_derivatives=True,
    )
    # No value assigned at 0 because there's no corresponding discretization equation
    assert m.dxdt[0].value is None
    assert pyo.value(m.dxdt[1]) == pytest.approx(-0.7580125427537873, rel=1e-3)
    assert m.dxdt[2].value is None


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_calculate_derivatives_integrate_second_half():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=2, scheme="BACKWARD")

    m.u.fix(1.0)
    m.x[1].fix(-0.7580125427537873)
    m.diff_eq[0].deactivate()

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[1.0, 2.0],
        ts_options={
            "--ts_type": "beuler",
            "--ts_dt": 3e-2,
        },
        skip_initial=True,  # With u and x fixed, no variables to solve for at t0
        calculate_derivatives=True,
    )
    # No value assigned at 0 because there's no corresponding discretization equation
    assert m.dxdt[0].value is None
    assert m.dxdt[1].value is None
    assert pyo.value(m.dxdt[2]) == pytest.approx(-0.2034733131369807, rel=1e-3)


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_calculate_derivatives_forward_difference():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=2, scheme="FORWARD")

    m.u.fix(1.0)
    m.x[0].fix(0.0)

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[0.0, 2.0],
        ts_options={
            "--ts_type": "beuler",
            "--ts_dt": 3e-2,
        },
        skip_initial=True,  # With u and x fixed, no variables to solve for at t0
        calculate_derivatives=True,
    )
    # It should backfill values for interpolated points like t=1
    assert pyo.value(m.dxdt[0]) == pytest.approx(-0.7580125427537873, rel=1e-3)
    assert pyo.value(m.dxdt[1]) == pytest.approx(-0.2034733131369807, rel=1e-3)
    # No value assigned at 2 because there's no corresponding discretization equation
    # TODO I don't understand how a value is getting assigned here
    # assert m.dxdt[2].value is None


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_calculate_derivatives_central_difference():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, nfe=2, scheme="CENTRAL")

    m.u.fix(1.0)
    m.x[0].fix(0.0)

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[0.0, 2.0],
        ts_options={
            "--ts_type": "beuler",
            "--ts_dt": 3e-2,
        },
        skip_initial=True,  # With u and x fixed, no variables to solve for at t0
        calculate_derivatives=True,
    )

    assert m.dxdt[0].value is None
    # It should backfill values for interpolated points like t=1
    assert pyo.value(m.dxdt[1]) == pytest.approx(-0.480742927945384, rel=1e-3)
    # No value assigned at 2 because there's no corresponding discretization equation
    # TODO I don't understand how a value is getting assigned here
    # assert m.dxdt[2].value is None


@pytest.mark.unit
@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
def test_calculate_derivatives_collocation_lr():
    m = pyo.ConcreteModel()

    m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
    m.x = pyo.Var(m.time)
    m.u = pyo.Var(m.time)
    m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

    def diff_eq_rule(m, t):
        return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

    m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

    discretizer = pyo.TransformationFactory("dae.collocation")
    discretizer.apply_to(m, nfe=2, scheme="LAGRANGE-RADAU")

    m.u.fix(1.0)
    m.x[0].fix(0.0)

    res = petsc.petsc_dae_by_time_element(
        m,
        time=m.time,
        between=[0.0, 2.0],
        ts_options={
            "--ts_type": "beuler",
            "--ts_dt": 3e-2,
        },
        skip_initial=True,  # With u and x fixed, no variables to solve for at t0
        calculate_derivatives=True,
    )

    assert m.dxdt[0].value is None
    # It should backfill values for interpolated points like t=1
    assert pyo.value(m.dxdt[1]) == pytest.approx(-0.3787483909827891, rel=1e-3)
    assert pyo.value(m.dxdt[2]) == pytest.approx(-0.07952012649572815, rel=1e-3)


# This test fails because the PETSc interface doesn't work with Lagrange-Legendre collocation
# @pytest.mark.unit
# @pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
# def test_calculate_derivatives_collocation_ll():
#     m = pyo.ConcreteModel()

#     m.time = pyodae.ContinuousSet(initialize=(0.0, 1.0, 2.0))
#     m.x = pyo.Var(m.time)
#     m.u = pyo.Var(m.time)
#     m.dxdt = pyodae.DerivativeVar(m.x, wrt=m.time)

#     def diff_eq_rule(m, t):
#         return m.dxdt[t] == m.x[t] ** 2 - m.u[t]

#     m.diff_eq = pyo.Constraint(m.time, rule=diff_eq_rule)

#     discretizer = pyo.TransformationFactory("dae.collocation")
#     discretizer.apply_to(m, nfe=2, scheme="LAGRANGE-LEGENDRE")

#     m.u.fix(1.0)
#     m.x[0].fix(0.0)

#     res = petsc.petsc_dae_by_time_element(
#         m,
#         time=m.time,
#         between=[0.0, 2.0],
#         ts_options={
#             "--ts_type": "beuler",
#             "--ts_dt": 3e-2,
#         },
#         skip_initial=True,  # With u and x fixed, no variables to solve for at t0
#         calculate_derivatives=True,
#     )

#     assert m.dxdt[0].value is None
#     # It should backfill values for interpolated points like t=1
#     assert pyo.value(m.dxdt[1]) == pytest.approx(-0.3787483909827891, rel=1e-3)
#     assert pyo.value(m.dxdt[2]) == pytest.approx(-0.07952012649572815, rel=1e-3)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Tests for the dynamic model subclass of block
"""

import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet
from pyomo.core.expr.visitor import identify_variables
from pyomo.core.base.block import BlockData, SubclassOf
from pyomo.dae.flatten import flatten_dae_components

from idaes.apps.caprese.tests.test_simple_model import (
    make_model,
    make_small_model,
    initialize_t0,
    copy_values_forward,
)
from idaes.apps.caprese.dynamic_block import (
    DynamicBlock,
    SimpleDynamicBlock,
    IndexedDynamicBlock,
    _DynamicBlockData,
)
from idaes.apps.caprese.categorize import categorize_dae_variables
from idaes.apps.caprese.common.config import (
    VariableCategory,
    InputOption,
)
from idaes.core.solvers import get_solver

VC = VariableCategory
from idaes.apps.caprese.nmpc_var import (
    NmpcVar,
    DiffVar,
    DerivVar,
    AlgVar,
    InputVar,
    FixedVar,
)
import pytest

__author__ = "Robert Parker"

solver_available = pyo.SolverFactory("ipopt").available()
if solver_available:
    solver = get_solver(solver="ipopt")
else:
    solver = None


class TestDynamicBlock(object):
    @pytest.mark.unit
    def test_init_simple(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        measurements = [model.conc[0, "A"], model.conc[0, "B"]]
        block = DynamicBlock(
            model=model,
            time=time,
            inputs=inputs,
            measurements=measurements,
        )
        # Assert that we have the correct type
        assert type(block) is SimpleDynamicBlock
        assert isinstance(block, DynamicBlock)
        assert isinstance(block, _DynamicBlockData)

        block.construct()
        # Assert that we behave like a simple block
        assert block[None] is block
        assert all(b is block for b in block[:])

        # Assert that input attributes have been processed correctly
        assert block.mod is model
        assert block.time is time
        assert all(i1 is i2 for i1, i2 in zip(block._inputs, inputs))
        assert all(i1 is i2 for i1, i2 in zip(block._measurements, measurements))

        # Assert that utility attributes have been added
        assert hasattr(block, "category_dict")
        assert hasattr(block, "vardata_map")
        assert hasattr(block, "measurement_vars")
        assert hasattr(block, "differential_vars")
        assert hasattr(block, "algebraic_vars")
        assert hasattr(block, "derivative_vars")
        assert hasattr(block, "input_vars")
        assert hasattr(block, "fixed_vars")

        subblocks = [
            block.mod,
            block.vectors,
            block.DIFFERENTIAL_BLOCK,
            block.ALGEBRAIC_BLOCK,
            block.INPUT_BLOCK,
            block.FIXED_BLOCK,
            block.DERIVATIVE_BLOCK,
            block.MEASUREMENT_BLOCK,
        ]

        block_objects = ComponentSet(
            block.component_objects(pyo.Block, descend_into=False)
        )
        # Assert that subblocks have been added
        assert len(subblocks) == len(block_objects)
        for b in subblocks:
            assert b in block_objects

        # Assert that we can add variables and constraints to the block
        block.v = pyo.Var(initialize=3)
        block.c = pyo.Constraint(expr=block.v == 5)
        assert block.v.value == 3
        assert block.v in ComponentSet(identify_variables(block.c.expr))

    @pytest.mark.unit
    def test_init_indexed(self):
        block_set = pyo.Set(initialize=[0, 1, 2])
        block_set.construct()
        horizon_map = {0: 1.0, 1: 3.0, 2: 5.0}
        nfe_map = {0: 2, 1: 6, 2: 10}
        model_map = {i: make_model(horizon_map[i], nfe_map[i]) for i in block_set}
        time_map = {i: model_map[i].time for i in block_set}
        inputs_map = {i: [model_map[i].flow_in[0]] for i in block_set}
        measurements_map = {
            i: [model_map[i].conc[0, "A"], model_map[i].conc[0, "B"]] for i in block_set
        }
        # Construct block with a dict for each of its arguments
        block = DynamicBlock(
            block_set,
            model=model_map,
            time=time_map,
            inputs=inputs_map,
            measurements=measurements_map,
        )
        # Make sure we have the right type
        assert type(block) is IndexedDynamicBlock
        assert isinstance(block, DynamicBlock)

        block.construct()
        assert all(b.parent_component() is block for b in block.values())

        # Check __contains__
        for i in block_set:
            assert i in block

        # Check attributes and subblocks of each data object
        for i, b in block.items():
            assert b.mod is model_map[i]
            assert b.time is time_map[i]
            assert all(i1 is i2 for i1, i2 in zip(b._inputs, inputs_map[i]))
            assert all(i1 is i2 for i1, i2 in zip(b._measurements, measurements_map[i]))

            assert hasattr(b, "category_dict")
            assert hasattr(b, "vardata_map")
            assert hasattr(b, "measurement_vars")
            assert hasattr(b, "differential_vars")
            assert hasattr(b, "algebraic_vars")
            assert hasattr(b, "derivative_vars")
            assert hasattr(b, "input_vars")
            assert hasattr(b, "fixed_vars")

            subblocks = [
                b.mod,
                b.vectors,
                b.DIFFERENTIAL_BLOCK,
                b.ALGEBRAIC_BLOCK,
                b.INPUT_BLOCK,
                b.FIXED_BLOCK,
                b.DERIVATIVE_BLOCK,
                b.MEASUREMENT_BLOCK,
            ]

            block_objects = ComponentSet(
                b.component_objects(pyo.Block, descend_into=False)
            )
            assert len(subblocks) == len(block_objects)
            for sb in subblocks:
                assert sb in block_objects

            b.v = pyo.Var(initialize=3)
            b.c = pyo.Constraint(expr=b.v == 5)
            assert b.v.value == 3
            assert b.v in ComponentSet(identify_variables(b.c.expr))

    @pytest.mark.unit
    def test_init_rule(self):
        block_set = pyo.Set(initialize=range(3))
        block_set.construct()
        # Create same maps as before
        horizon_map = {0: 1, 1: 3, 2: 5}
        nfe_map = {0: 2, 1: 6, 2: 10}
        model_map = {
            i: make_model(horizon=horizon_map[i], nfe=nfe_map[i]) for i in block_set
        }

        # Create rule to construct DynamicBlock with
        def dynamic_block_rule(b, i):
            model = model_map[i]
            time = model.time
            t0 = time.first()
            inputs = [model.flow_in[t0]]
            measurements = [model.conc[0, "A"], model.conc[0, "B"]]

            # Won't be obvious that these attrs need to be set if
            # constructing from a rule
            b.mod = model
            super(BlockData, b).__setattr__("time", time)
            b._inputs = inputs
            b._measurements = measurements

        # Create DynamicBlock from a rule
        block = DynamicBlock(block_set, rule=dynamic_block_rule)
        assert type(block) is IndexedDynamicBlock
        assert isinstance(block, DynamicBlock)

        block.construct()

        # Make sure iterating over block.values works as expected
        assert all(b.parent_component() is block for b in block.values())

        # Make sure __contains__ works
        for i in block_set:
            assert i in block

        # Assert correct attributes and subblocks
        for i, b in block.items():
            assert b.mod is model_map[i]
            assert b.time is model_map[i].time
            t0 = b.time.first()
            assert all(
                i1 is i2 for i1, i2 in zip(b._inputs, [model_map[i].flow_in[t0]])
            )
            assert all(
                i1 is i2
                for i1, i2 in zip(
                    b._measurements,
                    [model_map[i].conc[t0, "A"], model_map[i].conc[t0, "B"]],
                )
            )

            assert hasattr(b, "category_dict")
            assert hasattr(b, "vardata_map")
            assert hasattr(b, "measurement_vars")
            assert hasattr(b, "differential_vars")
            assert hasattr(b, "algebraic_vars")
            assert hasattr(b, "derivative_vars")
            assert hasattr(b, "input_vars")
            assert hasattr(b, "fixed_vars")

            subblocks = [
                b.mod,
                b.vectors,
                b.DIFFERENTIAL_BLOCK,
                b.ALGEBRAIC_BLOCK,
                b.INPUT_BLOCK,
                b.FIXED_BLOCK,
                b.DERIVATIVE_BLOCK,
                b.MEASUREMENT_BLOCK,
            ]

            block_objects = ComponentSet(
                b.component_objects(pyo.Block, descend_into=False)
            )
            assert len(subblocks) == len(block_objects)
            for sb in subblocks:
                assert sb in block_objects

            b.v = pyo.Var(initialize=3)
            b.c = pyo.Constraint(expr=b.v == 5)
            assert b.v.value == 3
            assert b.v in ComponentSet(identify_variables(b.c.expr))

    @pytest.mark.unit
    def test_construct(self):
        m = make_small_model()
        time = m.time
        t0 = time.first()
        helper = DynamicBlock(
            model=m,
            time=time,
            inputs=[m.flow_in[0]],
            measurements=[m.conc[0, "A"], m.conc[0, "B"]],
        )
        helper.construct()
        assert hasattr(helper, "category_dict")
        assert hasattr(helper, "vardata_map")
        assert hasattr(helper, "measurement_vars")
        assert hasattr(helper, "differential_vars")
        assert hasattr(helper, "algebraic_vars")
        assert hasattr(helper, "derivative_vars")
        assert hasattr(helper, "input_vars")
        assert hasattr(helper, "fixed_vars")

        # Make sure category dict contains variables we expect
        assert VariableCategory.DIFFERENTIAL in helper.category_dict
        assert VariableCategory.ALGEBRAIC in helper.category_dict
        assert VariableCategory.DERIVATIVE in helper.category_dict
        assert VariableCategory.INPUT in helper.category_dict
        assert VariableCategory.FIXED in helper.category_dict

        pred_diff_vars = ComponentSet(
            (
                m.conc[t0, "A"],
                m.conc[t0, "B"],
            )
        )
        diff_vars = list(helper.component_objects(DiffVar))
        assert len(pred_diff_vars) == len(diff_vars)
        for var in diff_vars:
            assert var[t0] in pred_diff_vars

        pred_alg_vars = ComponentSet(
            (
                m.flow_out[t0],
                m.rate[t0, "A"],
                m.rate[t0, "B"],
            )
        )
        alg_vars = list(helper.component_objects(AlgVar))
        assert len(pred_alg_vars) == len(alg_vars)
        for var in alg_vars:
            assert var[t0] in pred_alg_vars

        pred_input_vars = ComponentSet((m.flow_in[t0],))
        input_vars = list(helper.component_objects(InputVar))
        assert len(pred_input_vars) == len(input_vars)
        for var in input_vars:
            assert var[t0] in pred_input_vars

        pred_fixed_vars = ComponentSet(
            (
                m.conc_in[t0, "A"],
                m.conc_in[t0, "B"],
            )
        )
        fixed_vars = list(helper.component_objects(FixedVar))
        assert len(pred_fixed_vars) == len(fixed_vars)
        for var in fixed_vars:
            assert var[t0] in pred_fixed_vars

        pred_deriv_vars = ComponentSet(
            (
                m.dcdt[t0, "A"],
                m.dcdt[t0, "B"],
            )
        )
        deriv_vars = list(helper.component_objects(DerivVar))
        assert len(pred_deriv_vars) == len(deriv_vars)
        for var in deriv_vars:
            assert var[t0] in pred_deriv_vars

    @pytest.mark.unit
    def test_category_blocks(self):
        m = make_small_model()
        time = m.time
        t0 = time.first()
        helper = DynamicBlock(
            model=m,
            time=time,
            inputs=[m.flow_in[0]],
            measurements=[m.conc[0, "A"], m.conc[0, "B"]],
        )
        helper.construct()

        # Test that NmpcVectors and category blocks behave
        # as we expect
        diff_vars = helper.vectors.differential
        diff_var_set = helper.DIFFERENTIAL_SET * m.time
        assert diff_vars.index_set() == diff_var_set
        # And that they contain the same variables as the corresponding
        # lists on our helper block.
        for b, v in zip(helper.DIFFERENTIAL_BLOCK.values(), helper.differential_vars):
            assert b.var is v

        helper.vectors.derivative.set_setpoint(0.0)
        for var in helper.DERIVATIVE_BLOCK[:].var:
            assert var.setpoint == 0.0
        for b, v in zip(helper.DERIVATIVE_BLOCK.values(), helper.derivative_vars):
            assert b.var is v

        sp = [1, 2, 3]
        helper.vectors.algebraic.set_setpoint(sp)
        for i, j in zip(helper.ALGEBRAIC_SET, sp):
            assert helper.ALGEBRAIC_BLOCK[i].var.setpoint == j

        alg_vars = list(helper.component_objects(AlgVar))
        pred_alg_vars = ComponentSet(
            (
                m.flow_out[t0],
                m.rate[t0, "A"],
                m.rate[t0, "B"],
            )
        )
        assert len(pred_alg_vars) == len(alg_vars)
        for var in alg_vars:
            assert var[t0] in pred_alg_vars
        for b, v in zip(helper.ALGEBRAIC_BLOCK.values(), helper.algebraic_vars):
            assert b.var is v

        assert list(helper.vectors.algebraic.get_setpoint()) == sp

        vals = (10, 11)
        diff_vars.values = vals
        for var, val in zip(helper.DIFFERENTIAL_BLOCK[:].var, vals):
            for v in var[:]:
                assert v.value == val

        _slice = helper.DIFFERENTIAL_BLOCK[:].var
        diff_var_vals = diff_vars.values
        for var, vals in zip(_slice, diff_var_vals):
            for v, vl in zip(var[:], vals):
                assert v.value == vl

    @pytest.mark.unit
    def test_add_references(self):
        m = make_small_model()
        time = m.time
        t0 = time.first()
        blk = DynamicBlock(
            model=m,
            time=time,
            inputs=[m.flow_in[0]],
            measurements=[m.conc[0, "A"], m.conc[0, "B"]],
        )
        blk.construct()

        attrs = [
            "differential",
            "algebraic",
            "derivative",
            "input",
            "fixed",
            "measurement",
        ]

        for attr in attrs:
            # Make sure we've added the expected attribute
            assert hasattr(blk.vectors, attr)
            vec = getattr(blk.vectors, attr)

            # These "vectors" should be two-dimensional;
            # indexed by a coordinate (index into a list) and by time.
            assert vec.dim() == 2
            assert list(vec.index_set().subsets())[1] is time

            # Make sure we can use the set/get setpoint methods.
            # This should be tested more extensively elsewhere.
            setpoints = list(range(len(list(vec[:, t0]))))
            vec.set_setpoint(setpoints)
            assert list(vec.get_setpoint()) == setpoints

            # Make sure the underlying var has the ctype we expect
            # (check using the attribute/component name)
            for var in vec._generate_referenced_vars():
                assert var.ctype._attr == attr

    @pytest.mark.unit
    def test_validate_sample_time(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        measurements = [model.conc[t0, "A"], model.conc[t0, "B"]]
        blk = DynamicBlock(
            model=model,
            time=time,
            inputs=inputs,
            measurements=measurements,
        )
        blk.construct()
        blk.validate_sample_time(0.5)
        assert hasattr(blk, "sample_points")
        assert hasattr(blk, "fe_per_sample")
        assert hasattr(blk, "sample_point_indices")

        sample_point_set = set(blk.sample_points)
        sample_point_indices = set(blk.sample_point_indices)
        for p in [0.0, 0.5, 1.0]:
            assert p in sample_point_set
        for i in [1, 3, 5]:
            assert i in sample_point_indices
        assert len(sample_point_set) == 3
        assert len(sample_point_indices) == 3

        with pytest.raises(ValueError, match=r".*integer divider.*"):
            blk.validate_sample_time(0.6)

        with pytest.raises(ValueError, match=r"Could not find a time point.*"):
            blk.validate_sample_time(1 / 3.0)

        with pytest.raises(
            ValueError, match=r".*tolerance is larger than.*not.*unique.*"
        ):
            blk.validate_sample_time(0.5, tolerance=0.09)
            # min spacing in continuous set: 0.166667

        blk.validate_sample_time(0.5, tolerance=0.08)
        sample_point_set = set(blk.sample_points)
        for p in [0.0, 0.5, 1.0]:
            assert p in sample_point_set
        for i in [1, 3, 5]:
            assert i in sample_point_indices
        assert len(sample_point_set) == 3
        assert len(sample_point_indices) == 3

    @pytest.mark.unit
    def test_set_sample_time(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        measurements = [model.conc[t0, "A"], model.conc[t0, "B"]]
        blk = DynamicBlock(
            model=model,
            time=time,
            inputs=inputs,
            measurements=measurements,
        )
        blk.construct()

        blk.set_sample_time(1.0)
        assert blk.sample_points == [0.0, 1.0]

        blk.set_sample_time(0.5)
        assert blk.sample_points == [0.0, 0.5, 1.0]

    @pytest.mark.unit
    def make_block(self, sample_time=0.5, horizon=1, nfe=2):
        model = make_model(horizon=horizon, nfe=nfe)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        measurements = [model.conc[t0, "A"], model.conc[t0, "B"]]
        scalar_vars, dae_vars = flatten_dae_components(
            model,
            time,
            pyo.Var,
        )
        category_dict = categorize_dae_variables(
            dae_vars, time, inputs, measurements=measurements
        )
        dyn_block = DynamicBlock(
            model=model,
            time=time,
            category_dict={None: category_dict},
            # inputs=inputs,
            # measurements=measurements,
        )
        dyn_block.construct()
        dyn_block.set_sample_time(sample_time)
        return dyn_block

    @pytest.mark.unit
    def test_init_sample_to_setpoint(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()

        # Initialize all data objects with some consistent value
        # so I can make sure certain data objects are skipped
        # by the subsequent initialization
        init_val = -1.0
        blk.vectors.differential[...].set_value(init_val)
        blk.vectors.algebraic[...].set_value(init_val)
        blk.vectors.input[...].set_value(init_val)
        blk.vectors.derivative[...].set_value(init_val)

        # Note that I have no pointer from vectors.differential to
        # the list diff_vars... I can construct an identical list with
        # the _generate_referenced_vars method, but that is O(n)...
        # Not sure yet if this will be a problem/inconvenience...
        blk.vectors.differential.set_setpoint(range(len(blk.differential_vars)))
        blk.vectors.algebraic.set_setpoint(range(len(blk.algebraic_vars)))
        blk.vectors.input.set_setpoint(range(len(blk.input_vars)))
        blk.vectors.derivative.set_setpoint(range(len(blk.derivative_vars)))

        # Possible sample indices: {1, 2}
        sample_points = blk.sample_points
        blk.initialize_sample_to_setpoint(2)
        vectors = [
            blk.vectors.differential,
            blk.vectors.algebraic,
            blk.vectors.input,
            blk.vectors.derivative,
        ]
        for vec in vectors:
            for i, t in vec:
                if t <= sample_points[1]:
                    assert vec[i, t].value == init_val
                else:
                    assert vec[i, t].value == i

        blk.vectors.differential.set_setpoint(
            2 * i for i in range(len(blk.differential_vars))
        )
        blk.vectors.algebraic.set_setpoint(
            2 * i for i in range(len(blk.algebraic_vars))
        )
        blk.vectors.input.set_setpoint(2 * i for i in range(len(blk.input_vars)))
        blk.vectors.derivative.set_setpoint(
            2 * i for i in range(len(blk.derivative_vars))
        )

        blk.initialize_sample_to_setpoint(1)
        for vec in vectors:
            for i, t in vec:
                if t == t0:
                    assert vec[i, t].value == init_val
                elif t <= sample_points[1]:
                    assert vec[i, t].value == 2 * i
                else:
                    assert vec[i, t].value == i

    @pytest.mark.unit
    def test_init_to_setpoint(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()

        # Initialize all data objects with some consistent value
        # so I can make sure certain data objects are skipped
        # by the subsequent initialization
        init_val = -1.0
        blk.vectors.differential[...].set_value(init_val)
        blk.vectors.algebraic[...].set_value(init_val)
        blk.vectors.input[...].set_value(init_val)
        blk.vectors.derivative[...].set_value(init_val)

        # Note that I have no pointer from vectors.differential to
        # the list diff_vars... I can construct an identical list with
        # the _generate_referenced_vars method, but that is O(n)...
        # Not sure yet if this will be a problem/inconvenience...
        blk.vectors.differential.set_setpoint(range(len(blk.differential_vars)))
        blk.vectors.algebraic.set_setpoint(range(len(blk.algebraic_vars)))
        blk.vectors.input.set_setpoint(range(len(blk.input_vars)))
        blk.vectors.derivative.set_setpoint(range(len(blk.derivative_vars)))

        blk.initialize_to_setpoint()
        vectors = [
            blk.vectors.differential,
            blk.vectors.algebraic,
            blk.vectors.input,
            blk.vectors.derivative,
        ]
        for vec in vectors:
            for i, t in vec:
                if t == t0:
                    assert vec[i, t].value == init_val
                else:
                    assert vec[i, t].value == i

        new_sp = 7.0
        blk.vectors.algebraic.set_setpoint(new_sp)
        blk.initialize_to_setpoint(ctype=AlgVar)
        vectors = [
            blk.vectors.differential,
            blk.vectors.input,
            blk.vectors.derivative,
        ]
        for vec in vectors:
            for i, t in vec:
                if t == t0:
                    assert vec[i, t].value == init_val
                else:
                    assert vec[i, t].value == i

        vec = blk.vectors.algebraic
        for i, t in vec:
            if t == t0:
                assert vec[i, t].value == init_val
            else:
                assert vec[i, t].value == new_sp

    @pytest.mark.unit
    def test_init_sample_to_initial(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()

        # So I can tell when something wasn't initialized:
        init_val = -1.0  # `init_val` name might be confusing...
        blk.vectors.differential[...].set_value(init_val)
        blk.vectors.algebraic[...].set_value(init_val)
        blk.vectors.input[...].set_value(init_val)
        blk.vectors.derivative[...].set_value(init_val)

        vectors = [
            blk.vectors.differential,
            blk.vectors.algebraic,
            blk.vectors.derivative,
        ]
        # Set initial conditions to "index"
        for vec in vectors:
            for i, var in enumerate(vec[:, t0]):
                var.set_value(i)

        # Possible sample indices: {1, 2}
        sample_points = blk.sample_points
        blk.initialize_sample_to_initial(2)
        for vec in vectors:
            # We initialized sample 2 to the initial point in that
            # sample. This was just `init_val`
            for i, t in vec:
                if t != t0:
                    assert vec[i, t].value == init_val

        t1 = sample_points[1]
        for vec in vectors:
            for i, var in enumerate(vec[:, t1]):
                var.set_value(i)
        blk.initialize_sample_to_initial(2)
        for vec in vectors:
            for i, t in vec:
                if t == t0:
                    assert vec[i, t].value == i
                elif t < t1:
                    assert vec[i, t].value == init_val
                else:
                    assert vec[i, t].value == i

        blk.initialize_sample_to_initial(1)
        for vec in vectors:
            for i, t in vec:
                assert vec[i, t].value == i

    @pytest.mark.unit
    def test_init_to_initial_conditions(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()

        # So I can tell when something wasn't initialized:
        init_val = -1.0  # `init_val` name might be confusing...
        blk.vectors.differential[...].set_value(init_val)
        blk.vectors.algebraic[...].set_value(init_val)
        blk.vectors.input[...].set_value(init_val)
        blk.vectors.derivative[...].set_value(init_val)

        # Don't have a concise way to set value at t0
        # from a vector.
        # E.g. vectors.differential[:,t0].set_value(range(len(diff_vars)))
        # or vectors.differential[:,t0] = range(len(diff_vars))
        # The latter requires handling iterables in a slice setitem call
        vectors = [
            blk.vectors.differential,
            blk.vectors.algebraic,
            blk.vectors.input,
            blk.vectors.derivative,
        ]
        # Set initial conditions to "index"
        for vec in vectors:
            for i, var in enumerate(vec[:, t0]):
                var.set_value(i)

        blk.initialize_to_initial_conditions(ctype=(DiffVar, AlgVar))
        for vec in (blk.vectors.differential, blk.vectors.algebraic):
            for i, t in vec:
                assert vec[i, t].value == i
        for vec in (blk.vectors.input, blk.vectors.derivative):
            for i, t in vec:
                if t == t0:
                    assert vec[i, t].value == i
                else:
                    assert vec[i, t].value == init_val
                    # value that was used to initialize, not the initial
                    # condition value...

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason="IPOPT is not available")
    def test_init_by_solving_elements(self):
        blk = self.make_block()
        time = blk.time
        blk.mod.flow_in[:].set_value(3.0)
        initialize_t0(blk.mod)
        copy_values_forward(blk.mod)
        blk.mod.flow_in[:].set_value(2.0)
        blk.initialize_by_solving_elements(solver)

        t0 = time.first()
        tl = time.last()
        vectors = blk.vectors
        assert vectors.input[0, tl].value == 2.0
        assert vectors.differential[0, tl].value == pytest.approx(3.185595567867036)
        assert vectors.differential[1, tl].value == pytest.approx(1.1532474073395755)
        assert vectors.derivative[0, tl].value == pytest.approx(0.44321329639889284)
        assert vectors.derivative[1, tl].value == pytest.approx(0.8791007531878847)

        vectors.input.set_setpoint(3.0)
        blk.initialize_by_solving_elements(solver, input_option=InputOption.SETPOINT)
        for t in time:
            if t == t0:
                assert vectors.input[0, t].value == 2.0
            else:
                assert vectors.input[0, t].value == 3.0
        assert vectors.differential[0, tl].value == pytest.approx(3.7037037037037037)
        assert vectors.differential[1, tl].value == pytest.approx(1.0746896480968502)
        assert vectors.derivative[0, tl].value == pytest.approx(0.1851851851851849)
        assert vectors.derivative[1, tl].value == pytest.approx(0.47963475941315314)

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason="IPOPT is not available")
    def test_init_samples_by_element(self):
        blk = self.make_block()
        time = blk.time
        blk.mod.flow_in[:].set_value(3.0)
        initialize_t0(blk.mod)
        copy_values_forward(blk.mod)
        blk.mod.flow_in[:].set_value(2.0)
        blk.set_sample_time(0.5)
        blk.initialize_samples_by_element((1, 2), solver)

        t0 = time.first()
        tl = time.last()
        vectors = blk.vectors
        assert vectors.input[0, tl].value == 2.0
        assert vectors.differential[0, tl].value == pytest.approx(3.185595567867036)
        assert vectors.differential[1, tl].value == pytest.approx(1.1532474073395755)
        assert vectors.derivative[0, tl].value == pytest.approx(0.44321329639889284)
        assert vectors.derivative[1, tl].value == pytest.approx(0.8791007531878847)

        blk.vectors.input.set_setpoint(3.0)
        blk.initialize_samples_by_element(1, solver, input_option=InputOption.SETPOINT)
        blk.initialize_samples_by_element(2, solver, input_option=InputOption.SETPOINT)
        for t in time:
            if t == t0:
                assert vectors.input[0, t].value == 2.0
            else:
                assert vectors.input[0, t].value == 3.0
        assert vectors.differential[0, tl].value == pytest.approx(3.7037037037037037)
        assert vectors.differential[1, tl].value == pytest.approx(1.0746896480968502)
        assert vectors.derivative[0, tl].value == pytest.approx(0.1851851851851849)
        assert vectors.derivative[1, tl].value == pytest.approx(0.47963475941315314)

    @pytest.mark.unit
    def test_advance_by(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()
        tl = time.last()

        for t in time:
            blk.vectors.differential[:, t].set_value(t)
            blk.vectors.algebraic[:, t].set_value(t)
            blk.vectors.derivative[:, t].set_value(t)
            blk.vectors.input[:, t].set_value(t)
            blk.vectors.fixed[:, t].set_value(t)

        ctypes = (DiffVar, DerivVar, AlgVar, InputVar, FixedVar)

        shift = tl - t0  # 1.0, two samples
        blk.advance_by_time(shift)
        for t in time:
            if t == t0:
                for v in blk.component_objects(ctypes):
                    assert v[t].value == tl
            else:
                for v in blk.component_objects(ctypes):
                    assert v[t].value == t

        ctypes_to_shift = (DiffVar, DerivVar)
        ctypes_to_not_shift = (AlgVar, InputVar, FixedVar)
        shift = (tl - t0) / 2  # 0.5, one sample
        idx_of_last_sample = time.find_nearest_index(tl - shift)
        time_of_last_sample = time.at(idx_of_last_sample)
        blk.advance_by_time(shift, ctype=ctypes_to_shift)
        for t in time:
            if t <= shift:
                for v in blk.component_objects(ctypes_to_shift):
                    assert v[t].value == t + shift
            else:
                for v in blk.component_objects(ctypes_to_shift):
                    assert v[t].value == t
            if t == t0:
                for v in blk.component_objects(ctypes_to_not_shift):
                    assert v[t].value == tl
            else:
                for v in blk.component_objects(ctypes_to_not_shift):
                    assert v[t].value == t

    @pytest.mark.unit
    def test_advance_one_sample(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()
        tl = time.last()

        for t in time:
            blk.vectors.differential[:, t].set_value(t)
            blk.vectors.algebraic[:, t].set_value(t)
            blk.vectors.derivative[:, t].set_value(t)
            blk.vectors.input[:, t].set_value(t)
            blk.vectors.fixed[:, t].set_value(t)

        ctypes = (DiffVar, DerivVar, AlgVar, InputVar, FixedVar)

        shift = tl - t0  # 1.0, two samples
        blk.set_sample_time(shift)
        blk.advance_one_sample()
        for t in time:
            if t == t0:
                for v in blk.component_objects(ctypes):
                    assert v[t].value == tl
            else:
                for v in blk.component_objects(ctypes):
                    assert v[t].value == t

        ctypes_to_shift = (DiffVar, DerivVar)
        ctypes_to_not_shift = (AlgVar, InputVar, FixedVar)
        shift = (tl - t0) / 2  # 0.5, one sample
        idx_of_last_sample = time.find_nearest_index(tl - shift)
        time_of_last_sample = time.at(idx_of_last_sample)
        blk.set_sample_time(shift)
        blk.advance_one_sample(ctype=ctypes_to_shift)
        for t in time:
            if t <= shift:
                for v in blk.component_objects(ctypes_to_shift):
                    assert v[t].value == t + shift
            else:
                for v in blk.component_objects(ctypes_to_shift):
                    assert v[t].value == t
            if t == t0:
                for v in blk.component_objects(ctypes_to_not_shift):
                    assert v[t].value == tl
            else:
                for v in blk.component_objects(ctypes_to_not_shift):
                    assert v[t].value == t

    @pytest.mark.unit
    def test_generate_time_in_sample(self):
        blk = self.make_block()
        time = blk.time
        ts = blk.sample_points[1]
        i_s = time.find_nearest_index(ts)

        time_in_sample = list(blk.generate_time_in_sample(ts))
        i_prev = time.find_nearest_index(ts - blk.sample_time)
        pred_time_in_sample = [time.at(i) for i in range(i_prev + 1, i_s + 1)]
        assert time_in_sample == pred_time_in_sample

        time_in_sample = list(blk.generate_time_in_sample(ts, include_t0=True))
        pred_time_in_sample = [time.at(i) for i in range(i_prev, i_s + 1)]
        assert time_in_sample == pred_time_in_sample

        i_0 = 1
        t0 = time.at(i_0)
        ts = time.last()
        time_in_sample = list(blk.generate_time_in_sample(ts, t0=t0))
        pred_time_in_sample = list(t for t in time if t != t0)
        assert time_in_sample == pred_time_in_sample

    @pytest.mark.unit
    def test_get_data_from_sample(self):
        blk = self.make_block()
        time = blk.time
        ts = blk.sample_points[2]

        n_diff = len(blk.DIFFERENTIAL_SET)
        n_input = len(blk.INPUT_SET)
        n_alg = len(blk.ALGEBRAIC_SET)

        for (i, t), var in blk.vectors.differential.items():
            var.set_value(i * t)
        for (i, t), var in blk.vectors.input.items():
            var.set_value((n_diff + i) * t)
        for (i, t), var in blk.vectors.algebraic.items():
            var.set_value((n_diff + n_input + i) * t)

        # Test default variables with out including t0
        data = blk.get_data_from_sample(ts)
        # By default extract values from differential and input variables
        data_time = list(blk.generate_time_in_sample(ts))

        assert len(data) == n_diff + n_input
        for var in blk.component_objects((DiffVar, InputVar)):
            _slice = var.referent
            cuid = pyo.ComponentUID(_slice)
            assert cuid in data

        for i, b in blk.DIFFERENTIAL_BLOCK.items():
            cuid = pyo.ComponentUID(b.var.referent)
            values = list(b.var[t].value for t in data_time)
            assert values == data[cuid]
        for i, b in blk.INPUT_BLOCK.items():
            cuid = pyo.ComponentUID(b.var.referent)
            values = list(b.var[t].value for t in data_time)
            assert values == data[cuid]

        # Test default variables, including the data point at t0
        data = blk.get_data_from_sample(ts, include_t0=True)
        data_time = list(blk.generate_time_in_sample(ts, include_t0=True))
        assert len(data) == n_diff + n_input
        for var in blk.component_objects((DiffVar, InputVar)):
            _slice = var.referent
            cuid = pyo.ComponentUID(_slice)
            assert cuid in data
        for i, b in blk.DIFFERENTIAL_BLOCK.items():
            cuid = pyo.ComponentUID(b.var.referent)
            values = list(b.var[t].value for t in data_time)
            assert values == data[cuid]
        for i, b in blk.INPUT_BLOCK.items():
            cuid = pyo.ComponentUID(b.var.referent)
            values = list(b.var[t].value for t in data_time)
            assert values == data[cuid]

        data = blk.get_data_from_sample(ts, variables=(VariableCategory.ALGEBRAIC,))
        data_time = list(blk.generate_time_in_sample(ts))

        assert len(data) == n_alg
        for var in blk.component_objects(AlgVar):
            cuid = pyo.ComponentUID(var.referent)
            assert cuid in data

        for i, b in blk.ALGEBRAIC_BLOCK.items():
            cuid = pyo.ComponentUID(b.var.referent)
            values = list(b.var[t].value for t in data_time)
            assert values == data[cuid]

    @pytest.mark.unit
    def test_set_variance(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()

        variance_list = [
            (var[t0], 0.05) for var in blk.component_objects(SubclassOf(NmpcVar))
        ]
        blk.set_variance(variance_list)

        for var in blk.DIFFERENTIAL_BLOCK[:].var:
            assert var.variance == 0.05
        for var in blk.INPUT_BLOCK[:].var:
            assert var.variance == 0.05
        for var in blk.ALGEBRAIC_BLOCK[:].var:
            assert var.variance == 0.05
        for var in blk.MEASUREMENT_BLOCK[:].var:
            assert var.variance == 0.05
        for var in blk.FIXED_BLOCK[:].var:
            assert var.variance == 0.05
        for var in blk.DERIVATIVE_BLOCK[:].var:
            assert var.variance == 0.05

    @pytest.mark.unit
    def test_generate_inputs_at_time(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()
        t1 = time.at(2)
        vals = list(0.1 * i for i in blk.INPUT_SET)
        for var, val in zip(blk.vectors.input[:, t1], vals):
            var.set_value(val)
        inputs = list(blk.generate_inputs_at_time(t1))
        assert inputs == vals

    @pytest.mark.unit
    def test_generate_measurements_at_time(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()
        ts = blk.sample_points[1]
        vals = list(0.1 * i for i in blk.MEASUREMENT_SET)
        for var, val in zip(blk.vectors.measurement[:, ts], vals):
            var.set_value(val)
        meas = list(blk.generate_measurements_at_time(ts))
        assert meas == vals

    @pytest.mark.unit
    def test_inject_inputs(self):
        blk = self.make_block()
        time = blk.time
        vals = list(0.25 * i for i in blk.INPUT_SET)
        blk.inject_inputs(vals)
        for b, val in zip(blk.INPUT_BLOCK.values(), vals):
            for t in time:
                assert b.var[t].value == val

    @pytest.mark.unit
    def test_load_measurements(self):
        blk = self.make_block()
        time = blk.time
        t0 = time.first()
        vals = list(0.25 * i for i in blk.MEASUREMENT_SET)
        blk.load_measurements(vals)
        for b, val in zip(blk.MEASUREMENT_BLOCK.values(), vals):
            assert b.var[t0].value == val

    @pytest.mark.unit
    def test_categories_only_measurement_input(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in]
        measurements = [
            pyo.Reference(model.conc[:, "A"]),
            pyo.Reference(model.conc[:, "B"]),
        ]
        category_dict = {
            VC.INPUT: inputs,
            VC.MEASUREMENT: measurements,
        }
        db = DynamicBlock(
            model=model,
            time=time,
            category_dict={None: category_dict},
        )
        db.construct()
        db.set_sample_time(0.5)

        db.vectors.input[:, t0].set_value(1.1)
        db.vectors.measurement[:, t0].set_value(2.2)
        assert model.flow_in[t0].value == 1.1
        assert model.conc[t0, "A"].value == 2.2
        assert model.conc[t0, "B"].value == 2.2

    @pytest.mark.component
    def test_initialize_only_measurement_input(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in]
        measurements = [
            pyo.Reference(model.conc[:, "A"]),
            pyo.Reference(model.conc[:, "B"]),
        ]
        category_dict = {
            VC.INPUT: inputs,
            VC.MEASUREMENT: measurements,
        }
        db = DynamicBlock(
            model=model,
            time=time,
            category_dict={None: category_dict},
        )
        db.construct()
        db.set_sample_time(0.5)

        db.mod.flow_in[:].set_value(3.0)
        initialize_t0(db.mod)
        copy_values_forward(db.mod)
        db.mod.flow_in[:].set_value(2.0)

        # Don't need to know any of the special categories to initialize
        # by element. This is only because we have an implicit discretization.
        db.initialize_by_solving_elements(solver)

        t0 = time.first()
        tl = time.last()
        vectors = db.vectors
        assert vectors.input[0, tl].value == 2.0
        assert vectors.measurement[0, tl].value == pytest.approx(3.185595567867036)
        assert vectors.measurement[1, tl].value == pytest.approx(1.1532474073395755)
        assert model.dcdt[tl, "A"].value == pytest.approx(0.44321329639889284)
        assert model.dcdt[tl, "B"].value == pytest.approx(0.8791007531878847)

    @pytest.mark.unit
    def test_extra_category(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in]
        measurements = [
            pyo.Reference(model.conc[:, "A"]),
            pyo.Reference(model.conc[:, "B"]),
        ]
        disturbances = [
            pyo.Reference(model.conc_in[:, "A"]),
            pyo.Reference(model.conc_in[:, "B"]),
        ]
        category_dict = {
            VC.INPUT: inputs,
            VC.MEASUREMENT: measurements,
            VC.DISTURBANCE: disturbances,
        }
        db = DynamicBlock(
            model=model,
            time=time,
            category_dict={None: category_dict},
        )
        db.construct()
        db.set_sample_time(0.5)

        db.vectors.input[:, t0].set_value(1.1)
        db.vectors.measurement[:, t0].set_value(2.2)
        db.vectors.disturbance[:, t0].set_value(3.3)
        assert model.flow_in[t0].value == 1.1
        assert model.conc[t0, "A"].value == 2.2
        assert model.conc[t0, "B"].value == 2.2
        assert model.conc_in[t0, "A"].value == 3.3
        assert model.conc_in[t0, "B"].value == 3.3

    @pytest.mark.unit
    def test_empty_category(self):
        model = make_model(horizon=1, nfe=2)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in]
        measurements = [
            pyo.Reference(model.conc[:, "A"]),
            pyo.Reference(model.conc[:, "B"]),
        ]
        category_dict = {
            VC.INPUT: inputs,
            VC.MEASUREMENT: measurements,
            VC.ALGEBRAIC: [],
        }
        db = DynamicBlock(
            model=model,
            time=time,
            category_dict={None: category_dict},
        )
        db.construct()
        db.set_sample_time(0.5)

        # Categories with no variables are removed.
        # If they were retained, trying to iterate
        # over, e.g., vectors.algebraic[:, :] would
        # fail due to inconsistent dimension.
        assert VC.ALGEBRAIC not in db.category_dict



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Simplified Flash Unit Model, only for IAPWS with mixed state.
Phase separator - inlet water/steam mixture is separated into liquid and vapor
streams.

Created: August 21 2020
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.environ import value

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
from idaes.core.util.tables import create_stream_table_dataframe

__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa, A. Lee)"


@declare_process_block_class("HelmPhaseSeparator")
class WaterFlashData(UnitModelBlockData):
    """
    Simplified Flash Unit Model Class, only for IAPWS with mixed state
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Dynamic model flag",
            doc="""Indicates whether the model is dynamic""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Args:
            None

        Returns:
            None
        """
        super(WaterFlashData, self).build()

        self.mixed_state = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )

        self.add_port("inlet", self.mixed_state)

        self.vap_state = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )

        self.liq_state = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )

        self.add_port("vap_outlet", self.vap_state)
        self.add_port("liq_outlet", self.liq_state)

        # vapor outlet state
        @self.Constraint(self.flowsheet().time)
        def vap_material_balance(b, t):
            return (
                1e-4 * b.mixed_state[t].flow_mol * b.mixed_state[t].vapor_frac
                == b.vap_state[t].flow_mol * 1e-4
            )

        @self.Constraint(self.flowsheet().time)
        def vap_enthalpy_balance(b, t):
            return (
                b.mixed_state[t].enth_mol_phase["Vap"] * 1e-4
                == b.vap_state[t].enth_mol * 1e-4
            )

        @self.Constraint(self.flowsheet().time)
        def vap_pressure_balance(b, t):
            return b.mixed_state[t].pressure * 1e-6 == b.vap_state[t].pressure * 1e-6

        # liquid outlet state
        @self.Constraint(self.flowsheet().time)
        def liq_material_balance(b, t):
            return (
                1e-4 * b.mixed_state[t].flow_mol * (1 - b.mixed_state[t].vapor_frac)
                == b.liq_state[t].flow_mol * 1e-4
            )

        @self.Constraint(self.flowsheet().time)
        def liq_enthalpy_balance(b, t):
            return (
                1e-4 * b.mixed_state[t].enth_mol_phase["Liq"]
                == b.liq_state[t].enth_mol * 1e-4
            )

        @self.Constraint(self.flowsheet().time)
        def liq_pressure_balance(b, t):
            return b.mixed_state[t].pressure * 1e-6 == b.liq_state[t].pressure * 1e-6

    def initialize_build(
        blk,
        state_args_water_steam=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Drum initialization routine.

        Keyword Arguments:
            state_args_water_steam : a dict of arguments to be passed to the
                           property package(s) for the control_volume of the
                           model to provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")

        init_log.info_low("Starting initialization...")
        # fix FeedWater Inlet
        flags_fw = fix_state_vars(blk.mixed_state, state_args_water_steam)
        blk.mixed_state.initialize(solver=solver, optarg=optarg, outlvl=outlvl)
        # initialize outlet states
        for t in blk.flowsheet().time:
            blk.vap_state[t].flow_mol = value(
                blk.mixed_state[t].flow_mol * blk.mixed_state[t].vapor_frac
            )
            blk.vap_state[t].enth_mol = value(blk.mixed_state[t].enth_mol_phase["Vap"])
            blk.vap_state[t].pressure = value(blk.mixed_state[t].pressure)
            blk.vap_state[t].vapor_frac = 1
            blk.liq_state[t].flow_mol = value(
                blk.mixed_state[t].flow_mol * (1 - blk.mixed_state[t].vapor_frac)
            )
            blk.liq_state[t].enth_mol = value(blk.mixed_state[t].enth_mol_phase["Liq"])
            blk.liq_state[t].pressure = value(blk.mixed_state[t].pressure)
            blk.liq_state[t].vapor_frac = 0
        # unfix variables
        revert_state_vars(blk.mixed_state, flags_fw)
        init_log.info_low("Initialization Complete.")

    def set_initial_condition(self):
        pass

    def calculate_scaling_factors(self):
        pass

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Inlet": self.inlet,
                "Liquid Outlet": self.liq_outlet,
                "Vapor Outlet": self.vap_outlet,
            },
            time_point=time_point,
        )
